/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   lib/maus.c: Allgemeine Hilfsfunktionen fuer den MausTausch.
*/
/* brainded AIX needs this at the top of the file */
#ifdef _AIX
 #pragma alloca
#endif

#include "mausconsys.h"

#include <stdio.h>
#include <syslog.h>
#include <assert.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "maus.h"
#include "domains.h"

#include "msg_io.h"

#include "groups.h"

#define LINEEND "\r\n"
#define MAUS_STRLEN	(256*2+2)

static void maus_mk_ref (metamsg_t * msg, char *mref, char *realref);
static size_t export_part (FILE * f, metamsg_t * meta, long partnr);

#if 0
static void add_ref (metalist_t ** r, char *s);
#endif

    /*
     * Eine Adresse im MausTauschformat (Uwe Ohse @ DU3) verwandeln in
     * RFC-konforme Adress- und Realnameteile, Uwe_Ohse@DU3.maus.de und
     * "Uwe Ohse".
     */
static BOOLEAN mausadr_to_rfcadr(
    char *mausadr, /* zu wandelnde Adresse */
    char *adresse, /* entstehende Adresspart */
    char *realname /* Realname, nur ber den Localpart */ );

/*
 * Eine Adresse im MausTauschformat ausgeben.
 * bernimmt die Rckwandlung von domainisierten Boxnamen 
 * (mausdu3.gun.de -> DU3) [notwendig weil MAUS zu dumm ist das selber
 * zu machen] und Fidoadressen [notwendig weil FIDO halt FIDO ist].
 */
static void maus_print_adr (FILE *f, char prefix, char *adr);

struct mt_private 
{
	BOOLEAN mayclose;
	BOOLEAN output;
	BOOLEAN eof;
	char *fname;
	FILE *f;

	BOOLEAN zurueckgestellt;
	size_t linesize;
	char *line;
};

static mio_retcode_t
mt_open (const char *host, const char *mode,
		 mio_t ** ret_handle, char *errbuf);
static mio_retcode_t mt_close(mio_t *chan);
static mio_retcode_t mt_export(mio_t *chan,metamsg_t *);
static mio_retcode_t mt_import(mio_t *chan,metamsg_t *);

mio_base_t mio_maustausch=
{
	"maustausch",

	mt_open,
	mt_import,
	mt_export,
	mt_close,

	0 /* usecount */
};


static mio_retcode_t
mt_open (const char *fname, const char *mode,
		 mio_t ** ret_handle, char *errbuf)
{
	BOOLEAN output = FALSE;
	struct mt_private *priv;
	mio_t *neu;

	if (!strcmp (mode, "w"))
		output = TRUE;
	else if (strcmp (mode, "r"))
	{
		sprintf (errbuf, "illegal mode %s for mt_open", mode);
		log (LOG_ERR, "Illegal mode %s for mt_open", mode);
		return MIO_ERROR;
	}

	neu = mio_alloc (&mio_maustausch);
	priv = malloc(sizeof(struct mt_private));
	if (priv)
		priv->fname=strdup(fname);
	if (!neu || !priv || !priv->fname)
	{
		free(neu);
		if (priv)
			free(priv->fname);
		free(priv);
		strcpy (errbuf, _("out of memory"));
		log (LOG_ERR, _("out of memory"));
		return MIO_RETRY;
	}

	priv->linesize=16384;
	priv->zurueckgestellt=FALSE;
	priv->eof=FALSE;
	priv->line=malloc(priv->linesize);
	priv->output=output;
	if (!priv->line)
	{
		free(priv->fname);
		free(priv);
		return MIO_RETRY;
	}
	if (!strcmp(fname,"-"))
	{
		if (output)
			priv->f=stdout;
		else
			priv->f=stdin;
		priv->mayclose=FALSE;
	}
	else
	{
		if (output)
			priv->f=fopen(fname,"w");
		else
			priv->f=fopen(fname,"r");
		if (!priv->f)
		{
			int errno1=errno;
			free(priv->fname);
			free(priv);
			free(neu);
			sprintf (errbuf, _("cannot open %s: %s"), fname,strerror(errno1));
			log (LOG_ERR, _("cannot open %s: %s"), fname,strerror(errno1));
			return MIO_RETRY;
		}
		priv->mayclose=TRUE;
	}
	neu->priv_data=priv;
	*ret_handle=neu;
	return MIO_OK;
}

static mio_retcode_t
mt_close(mio_t *chan)
{
	mio_retcode_t ok=MIO_OK;
	int errno1=0;
	struct mt_private *priv;

	assert(chan);

	priv=(struct mt_private *)chan->priv_data;

	assert(priv);

	if (!priv)
		return TRUE;
	if (priv->mayclose)
	{
		if (fclose(priv->f)!=0)
		{
			/* wir haben keine Mglichkeit, den Fehler an dieser Stelle
			 * nach oben zu melden */
			errno1=errno;
			log(LOG_ERR,_("cannot close %s: %s"),priv->fname,strerror(errno1));
			ok=MIO_ERROR;
		}
	}
	free(priv->line);
	free(priv->fname);
	free(priv);
	chan->priv_data=NULL;
	mio_free(chan);
	errno=errno1;
	return ok;
}

static void
maus_print_adr (FILE *f, char prefix, char *adr)
{
	box_t *box;
	char *p;
	char *q;

	p=strchr(adr,'@');

	if (!p)
		fprintf (f, "%c%s" LINEEND, prefix, adr);
	else {
		size_t l;
		fidodomain_t *fdom;
		q=p+1;
		while(isspace(*q)) /* yes, this should not happen */
			q++;
		l=strlen(q);
		if (opts.strip_domain) {
			box=dom_find_domain(q);
			if (box && box->kuerzel) {
				*p=0;
				fprintf(f,"%c%s@%s" LINEEND,prefix, adr,box->kuerzel);
				*p='@';
				return;
			}
		}

		if ((fdom=fido_find_idomain(q))!=NULL) {
			/* [p15].f604.n245.z2.xxx.yyy -> ZZZ 2:245/604.15 */
			char *copy;
			int point=0;
			int node;
			int net;
			int zone;
			size_t len;
			char *s;

			len=strlen(q)+1;
			copy=alloca(len);
			memcpy(copy,q,len);

			strlwr(copy);
			*strstr(copy,fdom->internet_domain)=0;
			s=copy;
			if (*s=='p') {
				point=(int) strtol(s+1,&s,10);
				if (*s++!='.') 
					goto no_fido;
			}
			if (*s!='f') 
				goto no_fido;
			node=(int) strtol(s+1,&s,10);
			if (*s++!='.' && *s!='n') 
				goto no_fido;
			net=(int) strtol(s+1,&s,10);
			if (*s++!='.' && *s!='z')
				goto no_fido;
			zone=(int) strtol(s+1,&s,10);
			*p=0;
			if (point)
				fprintf (f, "%c%s@%s %d:%d/%d.%d" LINEEND, prefix, adr, 
						fdom->fido_domain1,
						 zone,net,node,point);
			else
				fprintf (f, "%c%s@%s %d:%d/%d" LINEEND, prefix, adr, 
						fdom->fido_domain1,
						 zone,net,node);
			*p='@';
			return;

		}
	no_fido:
		fprintf (f, "%c%s" LINEEND, prefix, adr);
	}
}

static BOOLEAN
mausadr_to_rfcadr (char *mausadr, char *adress, char *realname)
{
	char domain[1024];
	char *p = mausadr;
	char *pud = realname;		/* Pointer User Domain */
	int in_user = 1;
	box_t *box;
	BOOLEAN ok=TRUE;

	domain[0] = '\0';

	/* Erster Teil: In Local- und Domainpart zerlegen */
	while (*p)
	{
		switch (*p)
		{
		case '@':
			*pud = '\0';
			pud = domain;
			in_user = 0;
			break;
		case ' ':
			/* mehrere Leerzeichen zusammenziehen */
			if (p[1] == ' ')
				break;
			if (p[1] != '@' && p[-1] != '@')
				*pud++ = *p;
			break;
		default:
			*pud++ = *p;
			break;
		}						/* switch *p */
		p++;
	}							/* while *p */
	if (pud)
		pud[0] = '\0';			/* String terminieren */
	/* 
	 * So, jetzt haben wir zwei Strings darin. Der erste ist der Localpart,
	 * mit Leerzeichen darin. Der zweite der Domainpart, mglicherweise 
	 * ebenfalls mit Leerzeichen darin, was auf Fido-hnliche Adressierung
	 * hindeutet.
	 */
	
	/* Wenn es ein Boxkrzel ist */
	if (!strpbrk(domain,".\t "))
	{
		char *kopie=entumlauten1(domain);
		XNULLCHECK(kopie);
		box=dom_find_boxkuerzel(kopie);
		free(kopie);
		if (box)
		{
			if (box->primdomain)
				strcpy(domain,box->primdomain);
			else
				strcat(domain,opts.mausdomain);
		}
	}
#if 0
	/*
	 * In jedem Fall wird der Domainpart entumlautet.
	 */
	convert_umlaute (domain);
#endif

	/* Fido-Style-Adresse? */
	if (strchr (domain, ' ') && strchr(domain,'/'))
	{
		/* Tja, "Fido ..." */
		char *dot_p;
		char *slash_p;
		char *colon_p;

		char help[1024];
		char *zone_p;
		fidodomain_t *f;
		const char *idomain;

		strcpy (help, domain);
		zone_p=strchr(help,' ');
		*zone_p++=0;
		while (isspace(*zone_p))
			zone_p++;
		f=fido_find_fdomain(help);
		if (!f)
		{
			/* Zweiter Versuch: Kennen wir wenigstens die Zone? */
			f=fido_find_zone(strtol(zone_p,NULL,10));
		}
		if (!f || !f->internet_domain)
		{
			/* die knnen wir nicht konvertieren. Was mit der Nachricht
			 * geschieht wird weiter oben entschieden - der Fehler
			 * ist mglicherweise nicht wichtig, und fidonet.org
			 * tut's auch. Oder die Nachricht ist eine PM und 
			 * Zustellung ist sehr wnschenswert, unabhngig vom
			 * Absender
			 */
			ok=FALSE;
			idomain="fidonet.org";
		}
		else
			idomain=f->internet_domain;
		*domain=0;
		dot_p = strchr (zone_p, '.');
		if (dot_p)
		{
			sprintf (domain, "p%s.", dot_p + 1);
			*dot_p = 0;
		}
		slash_p = strchr (zone_p, '/');
		if (slash_p)
		{
			sprintf (domain + strlen (domain), "f%s.", slash_p + 1);
			*slash_p = 0;
		}
		/* Fido 2:245/604.15 oder Fido 245/604.15 */
		colon_p = strchr (zone_p, ':');
		if (colon_p)
		{
			sprintf (domain + strlen (domain), "n%s.", colon_p + 1);
			*colon_p = 0;
			sprintf (domain + strlen (domain), "z%s.", zone_p);
		}
		else
		{
			/* Keine Zone angegeben ... */
			/* Fido 245/604.15 */
			sprintf (domain + strlen (domain), "n%s.z2.", zone_p);
		}
		sprintf (domain + strlen (domain), idomain);
	}							/* if "Fido"-Style Domain */

	/* 
	 * So, jetzt haben wir eine gltige Domain.
	 * Jetzt die Adresse zusammensetzen ...
	 */
	{
		char *p1,*p2;
		p1=entumlauten1(realname);
		XNULLCHECK(p1);
		p2=entumlauten1(domain);
		XNULLCHECK(p2);
		sprintf (adress, "%s@%s", p1, p2);
		free(p1);
		free(p2);
	}
	/* und Leerzeichen durch Spaces ... */
	p = adress;
	while (*p)
	{
		if (*p == ' ')
			*p = '_';
		p++;
	}

#if 0
	/* Eventuell fehlende Domain ersetzen */
	if (!strchr (domain, '.') && *domain)
		strcpy (adress + strlen (adress), opts.mausdomain);
#endif

	/* Ein Username ohne Spaces knnte angegeben worden sein, in dem
	 * Fall gibt es keinen Realnamen 
	 */
	if (!strchr (realname, ' '))
		realname[0] = '\0';
	return ok;
}


static char *
last_reference (metalist_t * r)
{
	if (!r)
		return NULL;
	while (r->next)
		r = r->next;
	return r->text;
}


static size_t
write_message_refs (FILE * f, metamsg_t * meta, metalist_t * ref_gruppe)
{
	char mref[MAUS_STRLEN];
	char rref[MAUS_STRLEN];
	char *ref;

	ref = last_reference (meta->references);
	if (!ref || !strchr (ref, '@'))
		return 0;
	maus_mk_ref (meta, mref, rref);

	if (opts.boxtyp==BOXTYP_QUARK)
		fprintf (f, "-%s" LINEEND, ref);
	else /* weniger fhige Boxtypen */
	{
		if (*mref)
		{
			/* 
			 * Eine mausinterne References konnte ermittelt werden 
			 * HACK ALARM: Abhngigkeit von A/P!
			 */
			if ( (*mref=='A' && !meta->is_mail)
					|| (*mref=='P' && meta->is_mail)
					|| (*mref!='P' && *mref!='A'))
				fprintf (f, "-%s" LINEEND, mref);
		}
		if (*rref)
		{
			/* Und eine echte ebenso */
			fprintf (f, "R%s" LINEEND, rref);
		}						/* Usermodus? */
	}
	return 0; /* Kein Bodyplatz belegt */
}

static void
write_message_ids (FILE * f, metamsg_t * meta, long partnr)
{
	if (opts.workmode == USER) {
		/* 
		 * Als Benutzer sind keine echten Message-IDs erlaubt, man mu
		 * temporre IDs vergeben. Die Mausprogrammierer halten das 
		 * auch noch fr sinnvoll, angeblich wre Chaos die Folge,
		 * drften Benutzer IDs selbst vergeben. Aber welchen Sinn 
		 * hat es, eine *eindeutige* Message-ID (und die erzeugen 
		 * Mail- und Newssysteme nunmal) wegzuwerfen, um eine
		 * neue, weniger eindeutige, zu erzeugen?
		 * IMO eine vllig berzogene Rcksichtnahme auf veraltete
		 * Systeme, zumindest fr die sicheren Systeme sollte da
		 * mehr erlaubt sein.
		 */
		/* Message-ID */
		if (partnr)
			fprintf (f, "#%s%ld" LINEEND, meta->transfer_id, partnr);
		else
			fprintf (f, "#%s" LINEEND, meta->transfer_id);

	}
	/* Usermodus? */ 
	else {
		char buf[512];
		/* t-online ist so schlecht ... es sendet Kommentare mit
		 * derselben ID zurck. Schliet die Gates! */
		if (meta->is_mail && opts.t_online_id_hack) {
			if (((meta->env_from && strstr (meta->env_from, "t-online.de"))
				 || (meta->head_from && strstr (meta->head_from->text, "t-online.de")))
				&& !strstr (meta->message_id, "t-online.de")) {
				char *s;
				s = concat (meta->message_id, ".t-online.hack", NULL);
				if (s) {
					free (meta->message_id);
					meta->message_id = s;
#if 0 /* kann das nicht ins MausNet senden wegen 16K-Grenzen-Problem und
         der Unfhigkeit, unbekannte Header durchzureichen. Bug! */
					meta_add_list (&meta->unknown_headers,
								   "X-Warning",
								   "id hacked by mauscon", 0);
#endif
				}				/* else hope for the best */
			}
		}
		strcpy (buf, meta->message_id);
		if (partnr) {
			/* [entsetzt aufkreisch] Hier folgt einer der widerlichsten
			 * Hacks des Programms: Das Veraendern der Message-ID. Notwendig
			 * ist das durch den sehr ungluecklichen Umstand, dass die
			 * Nachricht gesplittet werden musste (ein zweiter sehr
			 * abstossender Hack). 
			 * Verfahren: Es wird eine voellig ungueltige Topleveldomain
			 * "splitted-for-mausnet-#" angehaengt und die alte Domain
			 * mit %-Hack "korrigiert". Das Verfahren ist schmutzig, und
			 * wird nur durch die Tatsache sicher, dass am Ende der
			 * hinteren Teile der Nachricht (partnr!=0) noch eine D-Zeile
			 * erzeugt wird, die den Export in andere Netze verhindert.
			 * Zusaetzlich erhaelt auch der erste Teil eine D-Zeile, da er
			 * nicht vollstaendig und daher nicht exportierbar ist. 
			 * Nachrichten zu splitten ist eine Suende, aber die 
			 * eigentlichen Verbrecher sind die MAUS- und manche 
			 * Frontendprogrammierer.
			 */
			char *p;
			p = strrchr (buf, '@');
			if (p)
				*p = '%';
			sprintf (buf + strlen (buf), "@splitted-for-mausnet-part-%ld", partnr);
		}
		if (opts.boxtyp == BOXTYP_QUARK) {
			/*
			 * keine transfer-id -> wir arbeiten als Filter oder so.
			 */
			if (meta->transfer_id) {
				fprintf (f, "#%s" LINEEND, meta->transfer_id);
				fprintf (f, "I%s" LINEEND, buf);
			} else
				fprintf (f, "#%s" LINEEND, buf);
		} else {
			if (partnr)
				fprintf (f, "#%s%ld" LINEEND, meta->transfer_id, partnr);
			else
				fprintf (f, "#%s" LINEEND, meta->transfer_id);
			fprintf (f, "I%s" LINEEND, buf);
		}
	}							/* User- oder Gatemodus */
}

static size_t
export_part (FILE * f, metamsg_t * meta, long partnr)
{
	unsigned char zwi[1024];

	int is_crossposting = 0;
	int is_mail = 0;
	int is_mime = 0;
	size_t textsize = 0;
	int used_cludges = 0;
	char *ref;
	metalist_t *groups = NULL;
	char *m_content_type_for_body=NULL;
	BOOLEAN header_envelope_mismatch=FALSE;

	if (meta->references)
		ref = last_reference (meta->references);
	else
		ref = NULL;

	/* Einen Header plus eventuell Pseudoheader in die Datei schreiben */

	/*
	 * Die Message-ID schreiben.
	 */
	write_message_ids (f, meta, partnr);

	if (meta->is_mail)
	{
		/* Dann ist es eine Mail */
		char *primary=NULL;
		char *secondary=NULL;

		if (meta->env_to)
			primary=meta->env_to;
		if (meta->head_to)	
		{
			if (!primary)
				primary=meta->head_to->text;
			else
				secondary=meta->head_to->text;
		}

		is_mail = 1;

		if (primary) /* Anderes kommt hier nicht an */
		{
			maus_print_adr (f, 'A', primary);
			/* nun prfen wir, ob sich envelope- und Headerempfnger 
			 * deutlich unterscheiden. Falls ja, wird spter ein
			 * entsprecheder Vermerk in den Body gesetzt. Das dient
			 * der Entschrfung der Mailinglistenproblematik.
			 */
			if (secondary)
			{
				/* belster Code von Welt ... */
				char *p1=strchr(primary,'@');
				char *p2=strchr(secondary,'@');
				char *q1,*q2;
				size_t l1;
				size_t l2;

				/* Nein, das ist *nicht nur* Paranoia. Ich hab' schon 
				 * Adressen mit Spaces drin gesehen. Nicht da die 
				 * korrekt gewesen wren, aber aufgetreten sind sie 
				 */
				if (p1)
					q1=p1-1;
				else
					q1=primary+strlen(primary)-1;
				while (isspace(*q1))
					q1--;
				p1=primary;
				while (isspace(*p1))
					p1++;
				l1=q1-p1;

				if (p2)
					q2=p2-1;
				else
					q2=secondary + strlen(secondary)-1;
				while (isspace(*q2))
					q2--;
				p2=secondary;
				while (isspace(*p2))
					p2++;
				l2=q2-p2;

				if (l1)
				{
					if (l2!=l1)
						header_envelope_mismatch=TRUE;
					else if (0!=memcmp(p1,p2,l1))
						header_envelope_mismatch=TRUE;
				}
			}
		}

		write_message_refs (f, meta, groups);
	}
	else if (!meta->newsgroups)
	{
		/* ### LOG-Eintrag fehlt */
		return (size_t) - 1;
	}
	else
	{
		/* ffentliche Nachricht */
		metalist_t *ng;
		BOOLEAN grp_found=FALSE;

		ng = meta->newsgroups;
		/* Eine Gruppe kann auf jeden Fall gewandelt werden, das ist
		 * weiter oben schon geprft worden.
		 */
		while (ng)
		{
			grp_data_t *grp;
			grp = grp_suche_rfcname (ng->text,opts.pollsite);
			if (grp)
			{
				fprintf (f, "G%s" LINEEND, grp->mausname);
				grp_found=TRUE;
				if (!opts.send_crosspostings)
					break;
			}
			else if (opts.send_crosspostings)
				fprintf (f, "G%s" LINEEND, ng->text);
			ng = ng->next;
		}
		if (ng && ng->next)
			/* spter Textcludge erzeugen */
			is_crossposting = 1;
		if (!grp_found && !opts.send_crosspostings)
		{
			/* Das kann zwar in rfc2maus *hier* nicht passieren, aber 
			 * mglicherweise in irgendwelchen Filtern, die die
			 * Gruppenliste nicht laden.
			 */
			fprintf (f, "G%s" LINEEND, meta->newsgroups->text);
		}
		write_message_refs (f, meta, groups);

		if (meta->followup_to && opts.send_followup_to)
		{
			metalist_t *f_list_p;
			BOOLEAN found=FALSE;

			f_list_p = meta->followup_to;
			while (f_list_p)
			{
				if (!strcasecmp (f_list_p->text, "poster"))
				{
					fprintf (f, "Fposter"LINEEND);
					if (!opts.send_crosspostings)
						break;
				}
				else
				{
					grp_data_t *grp;
					grp=grp_suche_rfcname(f_list_p->text,opts.pollsite);
					if (grp)
					{
						fprintf (f, "F%s" LINEEND, grp->mausname);
						found=TRUE;
					}
					else if (opts.send_crosspostings)
					{
						fprintf (f, "F%s" LINEEND, f_list_p->text);
						found=TRUE;
					}
				}
				f_list_p = f_list_p->next;
			}
			/* Wenn berhaupt nichts bekannt war, dann eben so ... */
			if (!found)
				fprintf (f, "F%s" LINEEND, meta->followup_to->text);
		}
	}

	/* User drfen keine Absenderangabe verschicken */
	if (opts.workmode != USER)
	{
		/* Absender angegeben? */
		/* Just for the case */
		char *from=NULL;
		if (meta->head_from)
		{
			from=meta->head_from->text;
			if (meta->head_from->ext)
				fprintf (f, "N%s" LINEEND, meta->head_from->ext);
		}
		else if (meta->env_from && meta->is_mail)
		{
			from=meta->env_from;
			if (meta->head_from && meta->head_from->text 
				&& !strcasecmp(meta->head_from->text,meta->env_from))
			{
				if (meta->head_from->ext)
					fprintf (f, "N%s" LINEEND, meta->head_from->ext);
			}
		}
		else if (meta->sender)
		{
			from=meta->sender->text;
			if (meta->sender->ext)
				fprintf (f, "N%s" LINEEND, meta->sender->ext);
		}
		if (from)
			maus_print_adr(f,'V',from);
	}

	/* Datumsangabe erzeugen */
	if (meta->tm_date)
	{
		strftime (zwi, sizeof (zwi) - 1, "%Y%m%d%H%M", meta->tm_date);
		fprintf (f, "E%s" LINEEND, zwi);
	}

	/*
	 * Subject
	 */
	{
		size_t subjmax;			/* maximale Laenge des Subjects */
		char *p;

		if (meta->subject)
			strcpy (zwi, meta->subject);
		else
			strcpy (zwi, "Kein Betreff angegeben");

		subjmax = 250;

		/* Re: eventuell killen */
		if (opts.rekillmode == REKILL_NONE)
			/* nicht killen */
			p = zwi;
		else if (opts.rekillmode == REKILL_FULL)
		{
			/* immer killen */
			if (!strncasecmp (zwi, "re: ", 4))
				p = zwi + 4;
			else
				p = zwi;
		}
		else
			/* REKILL_SMART */
		{
			/* nur killen, wenn das Maximum berschritten wrde */
			if (!strncasecmp (zwi, "re: ", 4))
			{
				if (strlen (zwi) > subjmax)
					p = zwi + 4;
				else
					p = zwi;
			}
			else
				p = zwi;
		}
		/* Maus kann noch nicht mal selbst den Betreff krzen :-) */
		if (strlen (p) > subjmax && opts.boxtyp == BOXTYP_MAUS)
			p[subjmax] = '\0';
		fprintf (f, "W%s" LINEEND, p);
	}


	if (opts.workmode!=USER && meta->organization)
		fprintf (f, "O%s" LINEEND, meta->organization);

	if (opts.send_path && meta->path)
	{
		if (strlen(meta->path)>250)
		{
			char *p;
			p=meta->path+250;
			while (p>meta->path && *p!='!')
				p--;
			if (p>meta->path)
			{
				*p='\0';
				fprintf(f,"p%s"LINEEND,meta->path);
				*p='!';
			}
		}
		else
			fprintf(f,"p%s"LINEEND,meta->path);
	}
	if (meta->reply_to && opts.send_reply_to)
	{
		/* nur *ein* Reply-To, mehr ist wohl nicht sinnvoll, oder? */
		fprintf(f,"T%s"LINEEND,meta->reply_to->text);
	}
	if (meta->sender && opts.send_sender)
	{
		fprintf(f,"S%s"LINEEND,meta->sender->text);
	}

	/* Feststellen, ob die Nachricht MIME ist */
	if (meta->mime_version)
	{
		char *v=rfc_getword(meta->mime_version,NULL);
		if (v && 0==strcasecmp(v,"1.0"))
			is_mime=1;
		free(v);
	}
	if (!meta->content_type)
		meta->content_type=XSTRDUP("text/plain");

	if (is_mime)
	{
		/* wir mssen jetzt einen MIME-Header generieren */
		/* Format:
		 * M1.0; text/enriched; charset=...
		 * Problem dabei: charset ist leider zur Zeit nicht machbar. 
		 */
		char *type;
		char *next;
		char *line=NULL;
		size_t linesize;
		char *wpos;
		size_t l;

		is_mime=0;
		/* Typ holen. 1. Wort in der Zeile. */
		type=rfc_getword(meta->content_type,&next);
		if (!type)
			goto no_m_line;

		/* text/plain ohne Parameter behandeln wir gar nicht erst */
		if (0==strcasecmp(type,"text/plain") && !next)
		{
			free(type);
			goto no_m_line;
		}
		is_mime=1;

		linesize=1024;
		line=XMALLOC(linesize);

#define MIME_CONSTANT "M1.0; Content-Type: "
		memcpy(line,MIME_CONSTANT,sizeof(MIME_CONSTANT)-1);
		wpos=line+sizeof(MIME_CONSTANT)-1;
		l=strlen(type);
		if (l+(wpos-line) >= linesize)
		{
			size_t off=wpos-line;
			/* genug Platz schaffen */
			linesize=(linesize+l+1024) & (~1024);
			line=XREALLOC(line,linesize);
			wpos=line+off;
		}
		memcpy(wpos,type,l);
		free(type);
		wpos+=l;

		/* die folgende Schleife beseitigt unter anderem Kommentare */
		for (type=next ? rfc_getword(next,&next) : NULL;
			 type; type=next ? rfc_getword(next,&next) : NULL)
		{
			l=strlen(type);
			/* "; " == 2 */
			if (l+2+(wpos-line) >= linesize)
			{
				size_t off=wpos-line;
				/* genug Platz schaffen */
				linesize=(linesize+l+2+1024) & (~1024);
				line=XREALLOC(line,linesize);
				wpos=line+off;
			}
			*wpos++=';';
			*wpos++=' ';
			memcpy(wpos,type,l);
			free(type);
			wpos+=l;
		}
		*wpos=0;
		if (wpos-line>200)
		{
			/* Oh hilfe - das habe *ich* mit verschuldet (und ganz anders
			 * gemeint). */
			fputs("M1.0; Content-Type: obscure/anything; "
				  "really too many parameters", f);

			/* "M1.0; " berspringen */
			m_content_type_for_body=XSTRDUP(line+6);
		}
		else
			fputs(line,f);
		fputs(LINEEND,f);
	no_m_line:
		free(line);
	}
	/* ich hoffe an dieser Stelle irgendwann >-Zeilen ausgeben zu drfen */

	/* So, nun sind Textkludges auszugeben */

	/* Eventuell pate der Content-Type nicht in die Headerzeile */
	if (m_content_type_for_body)
	{
		textsize += fprintf (f, ":%s" LINEEND, m_content_type_for_body);
		free(m_content_type_for_body);
	}
	/* sonstige MIME-relevante Header ausgeben */
	if (meta->content_id)
		textsize += fprintf (f, ":Content-ID: %s" LINEEND, meta->content_id);
	if (meta->content_description)
		textsize += fprintf (f, ":Content-Description: %s" LINEEND, 
							 meta->content_description);

	if (meta->reply_to && !opts.send_reply_to)
	{
		/* nur *ein* Reply-To, mehr ist wohl nicht sinnvoll, oder? */
		textsize += fprintf (f, ":Reply-To: %s" LINEEND, meta->reply_to->text);
		used_cludges = 1;
	}

	if (meta->keywords)
	{
		textsize += fprintf (f, ":Keywords: %s" LINEEND, meta->keywords);
		used_cludges = 1;
	}

	if (meta->summary)
	{
		textsize += fprintf (f, ":Summary: %s" LINEEND, meta->summary);
		used_cludges = 1;
	}

	if (is_crossposting)
	{
		metalist_t *ng;

		textsize += fprintf (f, ":Newsgroups: ");
		ng = meta->newsgroups;
		while (ng)
		{
			grp_data_t *grp;
			grp=grp_suche_rfcname(ng->text,opts.pollsite);
			if (grp)
				textsize += fprintf (f, "%s", grp->mausname);
			else
				textsize += fprintf (f, "(%s)", ng->text);
			if (ng->next)
				textsize += fprintf (f, ",");
			else
				textsize += fprintf (f, LINEEND);
			ng = ng->next;
		}
		used_cludges++;
	}
	if (!is_mail && !opts.send_followup_to && meta->followup_to)
	{
		/* Bei News Followup-Cludge erzeugen */
		metalist_t *ng;

		textsize += fprintf (f, ":Followup-To: ");
		ng = meta->followup_to;
		while (ng)
		{
			if (!strcasecmp (ng->text, "poster"))
				textsize += fprintf (f, "poster");
			else
			{
				grp_data_t *grp;
				grp=grp_suche_rfcname(ng->text,opts.pollsite);
				if (grp)
					textsize += fprintf (f, "%s", grp->mausname);
				else
					textsize += fprintf (f, "(%s)", ng->text);
			}
			if (ng->next)
				textsize += fprintf (f, ",");
			else
				textsize += fprintf (f, LINEEND);
			ng = ng->next;
		}
		used_cludges++;
	}

	if (header_envelope_mismatch)
	{
		textsize += fprintf (f, ":X-To: %s" LINEEND,
				meta->head_to->text);
	}	

	if (is_mime)
	{
		textsize += fprintf (f, ":-" LINEEND);
	} 
	else if (used_cludges)
	{
		/* Dann eine Leerzeile nach den Pseudoheadern einbauen */
		textsize += fprintf (f, ":%s", LINEEND);
	}
	return textsize;
}

/* 
 * Die zentrale Exportroutine fr Nachrichten im MausTausch-Format.
 * Sie wird durch die Notwendigkeit, lngere Nachrichten zu splitten,
 * nicht unbedingt eleganter. 
 */
static mio_retcode_t 
mt_export (mio_t * para, metamsg_t * meta)
{
	struct mt_private *priv=(struct mt_private *)para->priv_data;
	FILE *f;
	char *start;
	int partnr;

	assert(priv);
	assert(meta);

	f=priv->f;

	start = meta->body.start;
	partnr = 0;

	while (start && *start)
	{
		size_t hlen;
		int llen;

		hlen = export_part (f, meta, partnr);
		/* Zuerst mal die relativ schne Variante ohne TAB-Wandelung. */
		if (!opts.tabsize)
		{
			long in_line=0;
			while (start && *start)
			{
				char *p;
				p=strpbrk(start,"\r\n");
				if (p)
					llen=p-start;
				else
					llen=strlen(start);

				/* Immer eine Zeile durchlassen */
				if (in_line++>0 && opts.split_after 
					&& llen+2+hlen > ((size_t) opts.split_after))
					break;
				fprintf (f, ":%*.*s" LINEEND,
						 llen, llen, start);
				hlen+=llen+2;
				if (!p)
					start=NULL;
				else
				{
					if (*p=='\r')
						p++;
					if (*p=='\n')
						p++;
					start=p;
				}
			}
		}
		else /* Variante mit TAB-Wandelung */
		{
			int in_line=0;
			while (start && *start)
			{
				char *crlf;
				char *q;
				int lc=0;


				crlf=strchr(start,'\n');
				if (crlf && crlf>start && crlf[-1]=='\r')
					crlf--; /* auf CR */
				for (q=start; (crlf) ? q<crlf : *q;q++)
				{
					lc++;
					if (*q=='\t')
						lc+=opts.tabsize-(lc%opts.tabsize);
				}
				/* lc enthaelt nun die Laenge des Ausgabestrings ohne \r\n */
#if 0
				fprintf(f,"%ld + %ld = %ld\n",
					(long) lc, (long) hlen, (long) lc+hlen+2);
#endif
				/* Immer eine Zeile durchlassen */
				if (in_line++>0 && opts.split_after 
					&& lc+hlen+2>opts.split_after)
					break;
				hlen+=lc+2;
				putc (':', f);

				lc=0;
				for (q=start; (crlf) ? q<crlf : *q;q++)
				{
					if (*q=='\t')
					{
						do
						{
							putc (' ', f);
							lc++;
						}
						while (lc % opts.tabsize);
					}
					else
					{
						putc (*q, f);
						lc++;
					}
				}
				fputs (LINEEND, f);
				if (!crlf)
					start=NULL;
				else
				{
					if (*crlf=='\r')
						start=crlf+2;
					else
						start=crlf+1;
				}
			}
		}
		/* Falls die Nachricht gesplittet werden mute, dann die
		 * Distribution setzen, um Dupes zu verhindern.
		 */
		if (!meta->is_mail)
		{
			if (opts.workmode==GATE && (partnr || (start && *start)))
			{
				fprintf (f, "DL"LINEEND);
			}
			else if (meta->distribution)
			{
				if (!strcasecmp (meta->distribution, opts.usenet_dist_mausnet))
					fprintf (f, "DM" LINEEND);
				else if (!strcasecmp (meta->distribution, opts.pollsite)
					|| !strcasecmp(meta->distribution,opts.usenet_dist_lokal))
					fprintf (f, "DL" LINEEND);
				else
					fprintf (f, "DN" LINEEND);
			}
		}
		partnr++;
	}							/* while noch etwas brig */
	if (meta->is_mail && opts.generate_empty_mail_body)
	{
		/* PMs mit wichtigem Betreff ohne Text *knnen* auftreten */
		if (!meta->body.start || !*meta->body.start)
		{
			export_part (f, meta, partnr);
			fprintf (f, ":[kein Text vorhanden -- mauscon]" LINEEND);
		}
	}
	if (fflush(f)==0)
	{
		msglog("# %s: -> maustausch",meta->message_id);
		return MIO_OK;
	}
	else
	{
		msglog("# %s: -> %s: %s",meta->message_id,	
			   _("maustauschexport gescheitert"), strerror(errno));
		return MIO_ERROR;
	}
}

static void
make_path (metamsg_t * m, const char *group_path_entry,
	const char *foreign_export_path, const char *abs_unk_path)
{
	/* Eine Path-Zeile zusammensetzen.
	 * Gegenueber frueheren MausConnectversionen stark vereinfacht. 
	 * Den Pfad einer Nachricht im MausNet zu rekonstruieren ist 
	 * ziemlich witzlos.
	 */

	char *n;

	/* Den Namen unserer Kiste und den der Pollmaus in den Pfad setzen. */
	n=concat(
		(abs_unk_path && *abs_unk_path) 
			? abs_unk_path : "",
		(abs_unk_path && *abs_unk_path) ? "!" : "",
		(foreign_export_path && *foreign_export_path) 
			? foreign_export_path : "",
		(foreign_export_path && *foreign_export_path) ? "!" : "",
		(group_path_entry && *group_path_entry) ? group_path_entry : "",
		(group_path_entry && *group_path_entry) ? "!" : "",
		(opts.path_entry && *opts.path_entry) ? opts.path_entry : "",
		(opts.path_entry && *opts.path_entry) ? "!" : "",
		opts.pollsitedomain,
		"!not-for-mail",
		NULL);
	XNULLCHECK(n);
	m->path = entumlauten1 (n);
	XNULLCHECK(m->path);
}


static mio_retcode_t
mt_import(mio_t *channel, metamsg_t *meta)
{
	struct mt_private *priv;

	int weiter;
	int ende_bei_doppelkreuz;
	int in_text_kludges = 0;	/* In header or pseudoheader? */
	int raus = 0;
	mio_retcode_t retcode = MIO_OK;
	char *line;
	unsigned int unknown_group_count=0;
	grp_data_t *one_unknown_group=NULL;
	/* eigentlich 256, aber wg. Umlautwandlung besser mehr ... */
	unsigned char zwi_fromrealname[MAUS_STRLEN];
	unsigned char zwi_fromadress[MAUS_STRLEN];
	unsigned char zwi_maus_id[MAUS_STRLEN];	/* # */
	unsigned char zwi_real_id[MAUS_STRLEN];	/* I */
	unsigned char zwi_maus_ref[MAUS_STRLEN];	/* - */
	unsigned char zwi_real_ref[MAUS_STRLEN];	/* R */
	unsigned char zwi_to[MAUS_STRLEN];		/* A */
	unsigned char zwi_date[MAUS_STRLEN];	/* E */
	unsigned char tmp[1024];
	unsigned char tmp2[1024];
	const char *group_path_entry=NULL;
	const char *foreign_export_path=NULL;
	BOOLEAN no_export=FALSE;

	int b_found = 0;
	int g_found = 0;
	int m_found = 0;

	BOOLEAN absender_unkonvertierbar=FALSE;
	const char *abs_unk_path=NULL;

	assert(channel);
	assert(meta);

	priv=(struct mt_private *) channel->priv_data;

	if (priv->eof)
		return MIO_EOF;

	/* char xxx[..] = "" ruft memset auf ... */
	*zwi_fromrealname=0;
	*zwi_fromadress=0;
	*zwi_maus_id=0;
	*zwi_real_id=0;
	*zwi_maus_ref=0;
	*zwi_real_ref=0;
	*zwi_to=0;
	*zwi_date=0;
	*tmp=0;
	*tmp2=0;

	line=priv->line;

	/* Erst mal die Nachricht auf einen definierten leeren Zustand setzen */
	meta_clean_message (meta);
	/* Zeilenweise lesen */
	if (priv->zurueckgestellt)
		weiter = 1;
	else
	{
		weiter = fgets (line, priv->linesize - 1, priv->f) != NULL;
		if (!weiter)
		{
			priv->eof=TRUE;
			retcode = MIO_EOF;		/* Ende des Files */
		}
	}

	priv->zurueckgestellt = FALSE;
	ende_bei_doppelkreuz = 0;

	while (weiter)
	{
		char *crlfp;
		size_t len; /* Vorsicht: Wird auch weiter unten noch gebraucht! */

		/* strpbrk ist verhltnismig teuer */
		len=strlen(line);
		crlfp=len+line-1;
		if (*crlfp=='\n')
		{
			*crlfp--=0;
			len--;
		}
		if (*crlfp=='\r')
		{
			*crlfp--=0;
			len--;
		}

		switch (line[0])
		{
		case '#':
			if (line[1] == '\0')
			{
				/* Eine Zeile mit nur '#' beendet das Outfile */
				priv->eof=TRUE;
				/* das passiert, wenn die erste Zeile gleich "#" ist */
				if (!zwi_maus_id)
					retcode=MIO_EOF;
				else
					retcode = MIO_OK;
				raus = 1;
				break;
			}
			if (ende_bei_doppelkreuz)
			{
				priv->zurueckgestellt = TRUE;
				retcode = MIO_OK;	/* noch nicht zuende */
				raus = 1;
				break;
			}
			ende_bei_doppelkreuz = 1;

			in_text_kludges = 1;

			if (strchr (line + 1, '@'))
			{
				/* Muss sein */
				entumlauten(line+1,zwi_maus_id);
				meta->maus_id = XSTRDUP (zwi_maus_id);
			}
			else
			{
				/* Spezialnachricht */
				unsigned char timezwi[1024];
				minfo_t *found;

				struct tm *tm;
				time_t now;

				now = time (NULL);
				/* ID erzeugen */
				sprintf (zwi_real_id, "%s%ld@%s", line + 1, (long) now,
						 opts.mysitedomain);
				/* Special_ID sichern */
				meta->maus_special_id = XSTRDUP (line + 1);

				tm = localtime (&now);

				strftime (timezwi, sizeof (timezwi), "%c", tm);

				/* Sender erzeugen */
				sprintf (tmp, "MausConnect@%s", opts.mysitedomain);
					/* Fehler ist nicht so wild */
				meta_add_list (&meta->sender, tmp, NULL, 0);

				/* "From" erzeugen */
				sprintf (zwi_fromrealname, "%s %s",
					 opts.boxtyp == BOXTYP_MAUS ? "MAUS" : "Quark",
					 opts.pollsite);
				sprintf (zwi_fromadress, "%s@%s", 
					 opts.boxtyp == BOXTYP_MAUS ? "maus" : "quark",
					 opts.pollsitedomain);

				found = info_find (line + 1);
				if (found)
				{
					/* Ein Infofile */

					/* "Subject" */
					if (!found->kurztext || !*(found->kurztext))
					{
						char zwi2[1024];

						if (line[1] == 'I')
							sprintf (zwi2, "MAUSINFO %s (%s)", line + 1,
									 timezwi);
						else
							sprintf (zwi2, "MAUSDATEN %s (%s)", line + 1,
									 timezwi);
						meta->subject = strdup (zwi2);
					}
					else
					{
						char zwi2[1024];

						sprintf (zwi2, "%s (%s)", found->kurztext, timezwi);
						meta->subject = strdup (zwi2);
					}
					XNULLCHECK (meta->subject);
				}
				else if (!strcmp (line + 1, "LOG"))
				{
					/* "Subject" */
					char zwi[1024];

					sprintf (zwi, "LOG %s", timezwi);
					meta->subject = strdup (zwi);
					XNULLCHECK (meta->subject);
				}
				else if (!strcmp (line, "#HEAD"))
				{
					/* Eigentlich berflssig, wird weiter oben in der
					 * Hierarchie sowieso verworfen. 
					 */
					char zwi[1024];

					sprintf (zwi, "HEAD %s", timezwi);
					meta->subject = strdup (zwi);
					XNULLCHECK (meta->subject);
				}
				else if (!strcmp (line, "#REN"))
				{
					/* Eigentlich berflssig, wird weiter oben in der
					 * Hierarchie sowieso verworfen. 
					 */
					char zwi[1024];

					sprintf (zwi, "REN %s", timezwi);
					meta->subject = strdup (zwi);
					XNULLCHECK (meta->subject);
				}
				/* unbekannte Bloecke */
				else
				{
					unsigned char zwi2[256];

					/* "Subject" */
					sprintf (zwi2, "MAUSDATEN %s (%s)", line + 1, timezwi);
					meta->subject = strdup (zwi2);
					XNULLCHECK (meta->subject);
				}
			}
			break;
		case '-':
			entumlauten(line+1,zwi_maus_ref);
			break;
		case 'A':				/* an */
			strcpy (zwi_to, line + 1);
			break;
		case 'B':
			b_found = 1;
			meta->maus_status = XSTRDUP(line+1);
			break;
		case 'D':
			if (line[1]=='L')
				meta->distribution=XSTRDUP(opts.usenet_dist_lokal);
			else if (line[1]=='M')
				meta->distribution=XSTRDUP(opts.usenet_dist_mausnet);
			else if (line[1]!='N')
				meta->distribution=XSTRDUP(line+1); /* unbekannt, was nun? */
			break;
		case 'E':
			{
				meta->tm_date = XMALLOC(sizeof(struct tm));
				mtzeit_to_tm(line + 1,meta->tm_date);
			}
			/* Fr Erzeugung der Message-ID sichern */
			if (!strchr (line, ' ') && len-1 >= 12)
				strcpy (zwi_date, line + 1);
			break;
		case 'G':				/* gruppe */
			{
				grp_data_t *grp;

				grp=grp_suche_mausname(line+1,opts.pollsite);
				if (!grp)
				{

					/* neue Gruppe, noch nicht in der Liste */
					grp_create_rfcname(line+1,tmp);
					one_unknown_group=XMALLOC(sizeof(grp_data_t));
					*one_unknown_group=leer_grp;
					one_unknown_group->mausname=XSTRDUP(line+1);
					one_unknown_group->rfcname=XSTRDUP(tmp);
					one_unknown_group->boxen=XSTRDUP(opts.pollsite);
					grp=one_unknown_group;
					unknown_group_count++;
				}
				no_export|=grp->no_export;
				if (!group_path_entry)
					group_path_entry=grp->path_entry;
				/* hinten an die Liste der Gruppen anhaengen */
				if (!meta_add_list (&meta->newsgroups, grp->rfcname, NULL, 0))
				{
					log(LOG_CRIT,"out of memory");
					exit(1);
				}
				g_found = 1;
			}
			break;
		case 'I':
			entumlauten(line+1,zwi_real_id);
			break;
		case 'M':
			/* MIME-Versionen != 1.0 knnen wir nicht. 
			 * Gibt es jetzt auch nicht.
			 */
			if (0==strncmp(line+1,"1.0;",4))
			{
				char *p;
				meta->mime_version=XSTRDUP("1.0");
				p=line+5;
				while(isspace(*p))
					p++;
				if (0==strncasecmp("Content-Type:",p,13))
				{
					p+=13;
					while(isspace(*p))
						p++;
					meta->content_type=XSTRDUP(p);
				}
				m_found=1;
			}
			break;
		case 'N':
			strcpy (zwi_fromrealname, line + 1);
			break;
		case 'O':
			meta->organization = strdup (line + 1);
			XNULLCHECK (meta->organization);
			break;
		case 'R':
			entumlauten(line+1,zwi_real_ref);
			break;
		case 'V':				/* Von */
			if (!meta->organization)
			{
				box_t *box;
				char *p;
				p=strchr(line+1,'@');
				if (p)
				{
					p++;
					while (isspace(*p))
						p++;
					box=dom_find_boxkuerzel(p);
					if (!box)
						box=dom_find_domain(p);
					if (box)
					{
						if (box->orga)
							meta->organization=XSTRDUP(box->orga);
					}
				}
			}
			if (!mausadr_to_rfcadr (line + 1, zwi_fromadress, 
				zwi_fromrealname[0] ? tmp2 : zwi_fromrealname))
			{
				log(LOG_WARNING,_("cannot convert %s"),line+1);
				meta_add_list(&meta->unknown_headers,
					"X-Warning", 
					_("found illegal V (From) header"), 0);
				meta_add_list(&meta->unknown_headers,
					"X-Original-From", 
					line + 1, 0);
				absender_unkonvertierbar=TRUE;
			}
			break;
		case 'W':
			meta->subject = strdup (line + 1);
			XNULLCHECK (meta->subject);
			break;
		case 'Y':
			meta->x_gateway = XSTRDUP (line + 1);
			break;
		case ':':
			/* 
			 * Textzeilen? Irrtum - auch Textkludges.
			 * Verfahren:
			 * Fall 1: wir bekamen eine M-Zeile.
			 *         Wir lesen alle bekannten Kludges und enden mit 
			 *         einer Zeile, die nur "-" enthlt oder keinen
			 *         Doppelpunkt oder spaces vor dem Doppelpunkt
			 */
			if (in_text_kludges)
			{
				char *inhalt;
				char *doppelpunkt;
				char *ws;
				/* mit m-Zeile sind die Textkludges "offiziell" :-) */
				if (m_found && !strcmp(line+1,"-"))
				{
					/* Ende der Kludges erreicht */
					in_text_kludges=0;
					goto text_verarbeitet;
				}
				doppelpunkt=strchr(line+1,':');
				if (!doppelpunkt 
					|| ((ws=strpbrk(line+1,"\t "))!=NULL && ws<doppelpunkt))
				{
					in_text_kludges=0;
					goto normal_text;
				}
				inhalt=doppelpunkt;
				*inhalt++=0;
				while (isspace(*inhalt))
					inhalt++;
				if (!*inhalt)
				{
					/* Headerzeile ohne Inhalt? Hmmm ... */
					if (m_found)
						goto text_verarbeitet;
					/* Ohne M-Zeile gehen wir mal auf die sichere Seite! */
					*doppelpunkt=':';
					in_text_kludges=0;
					goto normal_text;
				}
				if (m_found)
				{
					if (!strhalfcasecmp(line+1,"content-type"))
					{
						meta->content_type=XSTRDUP(inhalt);
						goto text_verarbeitet;
					}
					else if (!strhalfcasecmp(line+1,"content-transfer-encoding"))
					{
						meta->content_transfer_encoding=XSTRDUP(inhalt);
						goto text_verarbeitet;
					}
					else if (!strhalfcasecmp(line+1,"content-description"))
					{
						meta->content_description=XSTRDUP(inhalt);
						goto text_verarbeitet;
					}
					else if (!strhalfcasecmp(line+1,"content-id"))
					{
						meta->content_id=XSTRDUP(inhalt);
						goto text_verarbeitet;
					}
				}
				if (!strhalfcasecmp(line+1,"followup-to"))
				{
					char *p2;
					
					p2 = inhalt;
					while (*p2 && !isspace(*p2))
						p2++;
					if (*p2!='\0' && !strncmp (inhalt, "PM ", 3))
						sprintf (tmp, "poster");
					else if (!*p2 && !strcmp (inhalt, "PM"))
						sprintf (tmp, "poster");
					else
					{
						grp_data_t *grp;
						grp=grp_suche_mausname(inhalt,opts.pollsite);
						if (!grp)
						{
							/* neue Gruppe, noch nicht in der Liste. 
							 * wird nicht abgespeichert!
							 */
							grp_data_t leer=LEER_GRP;

							grp_create_rfcname(inhalt,tmp);
							grp=XMALLOC(sizeof(grp_data_t));
							*grp=leer;
							grp->mausname=XSTRDUP(inhalt);
							grp->rfcname=XSTRDUP(tmp);
							grp->boxen=XSTRDUP(opts.pollsite);
							grp->dont_save=1;
							grp_append_new(grp);
						}
						else
							strcpy(tmp,grp->rfcname);
					}
					if (!meta_add_list (&meta->followup_to, tmp, NULL, 0))
					{
						log(LOG_CRIT,_("out of memory"));
						exit(1);
					}
					goto text_verarbeitet;
				}
				else if (!strhalfcasecmp(line+1,"reply-to"))
				{
					/* ### bei unkonvertierbarem Reply-To X-Zeile schreiben */
					if (!mausadr_to_rfcadr (inhalt, tmp, tmp2))
					{
						meta_add_list(&meta->unknown_headers,
							"X-Warning", 
							_("found illegal Reply-To"), 0);
						meta_add_list(&meta->unknown_headers,
							"X-Original-Reply-To", 
							inhalt, 0);
					}
					else if (!meta_add_list (&meta->reply_to, tmp, tmp2, 0))
					{
						log(LOG_CRIT,_("out of memory"));
						exit(1);
					}
					goto text_verarbeitet;
				}
				else if (!strhalfcasecmp(line+1,"keywords"))
				{
					meta->keywords = strdup (inhalt);
					goto text_verarbeitet;
				}
				else if (!strhalfcasecmp(line+1,"summary-to"))
				{
					meta->summary = strdup (inhalt);
					goto text_verarbeitet;
				}
				/* wenn wir hier ankommen gibt es zwei Mglichkeiten:
				 * 1. die Nachricht ist MIME und wir kennen diesen
				 *    Header nicht (schlecht: wir knnen unbekannte
				 *    Header nicht durchschleifen. Der Code fehlt 
				 *    noch). -> lossage, aber wir machen weiter. 
				 * 2. das knnte ein normaler Textkluge sein, den wir
				 *    nicht kennen. Das sollte nicht passieren und
				 *    es ist unwahrscheinlich, da dies wirklich
				 *    kein irrtmlich erkennter Header ist. Anders
				 *    gesagt: Wahrscheinlich ist es soetwas wie:
				 *        Hallo: Welt ich hab' dich lieb.
				 *    Also behandeln wir das als Ende der Kludges.
				 */
				if (!m_found)
				{
					in_text_kludges = 0;
					*doppelpunkt=':';
					goto normal_text;
				}
			}
			else
			{
			normal_text:		
				if (len+3 < priv->linesize)
				{
					line[len++]='\r';
					line[len++]='\n';
					line[len]='\0';
					meta_add_to_body (meta, line + 1);
				}
				else
				{
					meta_add_to_body (meta, line + 1);
					meta_add_to_body (meta, "\r\n");
				}
			}
		text_verarbeitet:
			break;
		}						/* switch */
		if (raus)
			break;				/* JETZT - vor dem fgets! */
		weiter = fgets (line, priv->linesize - 1, priv->f) != NULL;
	}							/* while */

	if (ferror(priv->f))
		retcode=MIO_ERROR;

	if (unknown_group_count==1)
	{
		grp_append_new(one_unknown_group);
		log(LOG_INFO,_("new newsgroup %s -> %s"),	
			one_unknown_group->mausname,one_unknown_group->rfcname);
	}

	/* Was muss jetzt noch geschehen? */

	/* Der Absender muss nun erzeugt werden */
	{
		if (!meta_add_list (&meta->head_from, zwi_fromadress, zwi_fromrealname, 
			0))
		{
			log(LOG_CRIT,_("out of memory"));
			exit(1);
		}
		meta->env_from = XSTRDUP (zwi_fromadress);
	}

	/* Empfaenger auswerten */
	if (*zwi_to)
	{
		if (meta->newsgroups)
		{
			meta->x_to=entumlauten1(zwi_to);
			XNULLCHECK (meta->x_to);
		}
		else
		{
			/* ### X-Original-To: schreiben bei Fehlschlag */
			if (!mausadr_to_rfcadr (zwi_to, tmp, tmp2))
			{
				meta_add_list(&meta->unknown_headers,
					"X-Warning", 
					_("found illegal To: Header"), 0);
				meta_add_list(&meta->unknown_headers,
					"X-Original-To", 
					zwi_to, 0);
			}
			if (!meta_add_list (&meta->head_to, tmp, tmp2, 0))
			{
				log(LOG_CRIT,_("out of memory"));
				exit(1);
			}
		}
	}

	/* Bei PMs muss jetzt eine Envelope-Adresse erzeugt werden */
	if (meta->env_to == NULL
		&& meta->newsgroups == NULL
		&& meta->head_to)
	{
		if (opts.workmode == GATE)
			meta->env_to = strdup (meta->head_to->text);
		else
			meta->env_to = strdup (opts.mailuserdomain);
		XNULLCHECK (meta->env_to);
	}

	/* - Eine ID mu erzeugt werden. Die MAUS-internen IDs sind ja 
	 *   bekanntlich nicht eindeutig.
	 */
	if (*zwi_real_id)			/* I */
	{
		meta->message_id=entumlauten1(zwi_real_id);
		XNULLCHECK (meta->message_id);
	}
	else if (strchr (zwi_maus_id, '.')
		|| !(isupper(*zwi_maus_id) && isdigit(zwi_maus_id[1])))		/* # */
	{
		/* Auf irgendeinem Weg ist eine vollstndige ID in die # -Zeile
		 * gekommen. vielleicht ein use2maus oder Quark 2.x.
		 */
		meta->message_id=entumlauten1(zwi_maus_id);
		XNULLCHECK (meta->message_id);
	}
	else if (*zwi_date)
	{
		char buf[MAUS_STRLEN];
		char *p;

		sprintf (buf, "%s.%s%s", zwi_date, zwi_maus_id, opts.mausdomain);
		meta->message_id = entumlauten1 (buf);
		XNULLCHECK (meta->message_id);

		/* Selbst zusammengesetzte IDs nach lowercase wandeln */
		p = meta->message_id;
		while (*p)
		{
			*p = tolower (*p);
			p++;
		}
	}
	else
	{
		/* Schrott, Infofile */
		char buf[MAUS_STRLEN];

		sprintf (buf, "%s%s", zwi_maus_id, opts.mausdomain);
		meta->message_id = entumlauten1 (buf);
		XNULLCHECK (meta->message_id);
	}

	/* - Eine References mu erzeugt werden */
	if (*zwi_real_ref || (*zwi_maus_ref && strchr (zwi_maus_ref, '.')))
	{
		char *work;

		/* eine "echte" References ist gefunden */
		if (*zwi_real_ref)
			work = zwi_real_ref;
		else
			work = zwi_maus_ref;
		work = entumlauten1 (work);
		XNULLCHECK(work);
		meta_add_list (&meta->references, work, NULL, 0);
		free(work);
	}

	meta->is_mail = 0;
	if (b_found)
		meta->is_mail = 1;
	if (g_found)
		meta->is_mail = 0;

	if ((opts.foreign_export_kill || opts.foreign_export_path
		|| opts.foreign_export_distribution) && !meta->is_mail)
	{
		char *copy=XSTRDUP(meta->message_id);
		strlwr(copy);
		if (!strstr(copy,opts.mausdomain))
		{
			foreign_export_path=opts.foreign_export_path;
			if (opts.foreign_export_distribution)
			{
				free(meta->distribution);
				meta->distribution=XSTRDUP(opts.foreign_export_distribution);
			}
			if (opts.foreign_export_kill) {
				msglog("# %s: foreign export",meta->message_id);
				retcode=MIO_RETRY;
			}
		}
	}

	/* X-Format-Zeile */
	meta->x_format = strdup ("MausTausch");		/* Error doesn't matter */

	if (absender_unkonvertierbar)
	{
		if (meta->is_mail)
		{
			/* ### X-Original-From ... */
			meta_add_list(&meta->unknown_headers,
				"X-Warning", 
				_("unconvertable V (From) header"),
				0);
		}
		else if (opts.unconvertable_from_kill)
		{
			msglog("# %s: from unconvertable",meta->message_id);
			retcode=MIO_RETRY;
		}
		else if (opts.unconvertable_from_path)
			abs_unk_path=opts.unconvertable_from_path;
		else if (opts.unconvertable_from_distribution)
		{
			free(meta->distribution);
			meta->distribution=XSTRDUP(opts.unconvertable_from_distribution);
		}
	}

	/* Pfad zusammensetzen */
	make_path (meta,group_path_entry,foreign_export_path,abs_unk_path);

	if (!meta->is_mail)
	{
		/* 
		 * Versuchen, noch ein paar Daten zusammenzuklauben, den
		 * Gruppennamen zu wandeln.
		 */
		metalist_t *ng;
		BOOLEAN had_followup=FALSE;
		BOOLEAN explicit_mime_header_true=FALSE;
		BOOLEAN explicit_mime_header_false=FALSE;
		BOOLEAN explicit_mime_body_true=FALSE;
		BOOLEAN explicit_mime_body_false=FALSE;

		if (meta->followup_to)
			had_followup=TRUE;

		for(ng=meta->newsgroups;ng;ng=ng->next)
		{
			grp_data_t *grp;
			grp = grp_suche_rfcname (ng->text,opts.pollsite);
			if (grp)
			{
				if (grp->default_followup && !had_followup)
				{
					meta_add_list(&meta->followup_to,
						grp->default_followup,NULL,0);
				}
				if (grp->mime_headers==TRI_TRUE)
					explicit_mime_header_true=TRUE;
				else if (grp->mime_headers==TRI_FALSE)
					explicit_mime_header_false=TRUE;
				if (grp->mime_body==TRI_TRUE)
					explicit_mime_body_true=TRUE;
				else if (grp->mime_body==TRI_FALSE)
					explicit_mime_body_false=TRUE;
				if (grp->approve && !meta->approved)
					meta->approved=XSTRDUP(grp->approve);
			}
		}
		if (explicit_mime_header_true)
			meta->mime_code_headers=TRUE;
		else if (explicit_mime_header_false)
			meta->mime_code_headers=FALSE;
		else
			meta->mime_code_headers=opts.mime_coded_headers;
		if (explicit_mime_body_true)
			meta->mime_code_body=TRUE;
		else if (explicit_mime_body_false)
			meta->mime_code_body=FALSE;
		else
			meta->mime_code_body=opts.mime_coded_body;
	}
	else
	{
		meta->mime_code_headers=opts.mime_coded_headers;
		meta->mime_code_body=opts.mime_coded_body;
	}

	if (!meta->is_mail && opts.workmode==GATE && no_export) {
		msglog("# %s: not exported",meta->message_id);
		retcode=MIO_RETRY;
	}
	if (retcode == MIO_EOF || feof (priv->f))
		priv->zurueckgestellt = FALSE;
	return retcode;
}


static void
maus_mk_ref (metamsg_t * msg, char *mref, char *rref)
{
	char *mdomp;
	char *dot;
	char c;
	char *at;
	metalist_t *l;

	*rref = '\0';
	*mref = '\0';

	if (!msg->references)
		return;
	l = msg->references;
	while (l->next)
		l = l->next;

	at=strchr (l->text, '@');
	if (!at)
		return;					/* das ist Schrott, soetwas interessiert nicht ... */

	mdomp = strstr (l->text, opts.mausdomain);
	/*
	 * Die Abfrage auf mdomp < at ist notwendig. Es kam mal soetwas:
	 * MSGID_=3C199601281843.a39758=40wi2.maus.de@fidonet.org
	 */
	if (!mdomp || mdomp<at)
	{
		/* Ahem - keine Mausdomain angegeben */
		strcpy (rref, l->text);
		return;
	}
	*mdomp = '\0';

	/* Falls Punkt im Rest des Domainparts, dann keine Maus-ID */
	if (strchr (at, '.'))
	{
		/* keine legale ID fr das MausNet */
		*mdomp = *opts.mausdomain;
		strcpy (rref, l->text);
		return;
	}
	dot = strchr (l->text, '.');
	if (dot)
	{
		*dot = '\0';
		c = toupper (dot[1]);
		if (c == 'A' || c == 'P' || opts.boxtyp != BOXTYP_MAUS)
		{
			/* Schn! */
			char *p;

			strcpy (mref, dot + 1);

			p = mref;
			while (*p)
			{
				*p = toupper (*p);
				p++;
			}

			*dot = '.';
			*mdomp = *opts.mausdomain;
			/* Keine Realref angeben? Oder doch? */
			strcpy (rref, l->text);
			return;
		}
		else
		{
			/* Keine legale Mausinterne ID */
			*dot = '.';
			*mdomp = *opts.mausdomain;
			strcpy (rref, l->text);
			return;
		}
	}
	else
	{
		/* Kein Punkt angegeben */
		c = toupper (*l->text);
		if (c == 'A' || c == 'P' || opts.boxtyp != BOXTYP_MAUS)
		{
			strcpy (mref, l->text);
			*mdomp = *opts.mausdomain;
			/* Keine Realref angeben! */
			return;
		}
		else
		{
			/* Keine legale Mausinterne ID */
			*mdomp = *opts.mausdomain;
			strcpy (rref, l->text);
			return;
		}
	}
}

#if 0
static void
add_ref (metalist_t ** r, char *s)
{
	char *p;

	while (s && *s)
	{
		p = strchr (s, ' ');
		if (p)
			*p = '\0';
		meta_add_list (r, s, NULL, 0); /* Fehler ist nicht soo wild */
		if (p)
			*p = ' ';
		s = p;
		/* nur zur Sicherheit */
		while (s && *s == ' ')
			s++;
	}
}
#endif
