/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   groups.c: Gruppennamenshandling in MausConnect.

*/

#include "mausconsys.h"

#include <stdio.h>
#include <assert.h>
#include <syslog.h>
#include <alloca.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "groups.h"
#include "dynhash.h"
#include "filelock.h"

static grp_data_t *anker = NULL;
static grp_data_t *last = NULL;
static FILE *groups_file;

grp_data_t leer_grp;

typedef struct hash_list_t 
{
	grp_data_t *data;
	struct hash_list_t *next;
} hash_list_t;

static void *maus_handle;
static void *rfc_handle;

static void canon_grp(const char *, char *);

static void
canon_grp(const char *from, char *to)
{
	const unsigned char *p=from;
	while (*p)
	{
		switch(*p)
		{
		case '':
		case '':
			*to++='u';
			*to++='e';
			break;
		case '':
		case '':
			*to++='o';
			*to++='e';
			break;
		case '':
		case '':
			*to++='a';
			*to++='e';
			break;
		case '':
			*to++='s';
			*to++='s';
			break;
		default:
			*to++=tolower(*p);
			break;
		}
		p++;
	}
	*to=0;
}

grp_data_t * 
grp_get_anker(void)
{
	return anker;
}

/*
 * Die Gruppennamenzuordnungsdatei einlesen. 
 * Format:
 * boxkuerzel<whitespace>name_in_rfc<whitespace>name_in_maus
 * Alles ab "#" ist Kommentar
 */
BOOLEAN
grp_load_list (BOOLEAN keep_locked)
{
	char fname[PATH_MAX];
	const char *mode="r";

	char *p;
	grp_data_t g=LEER_GRP;

	char str[1024];				/* Puffer fr eine gelesene Zeile */
	grp_data_t *new;
	long linenumber = 0;
	int ret;

	rfc_handle=hhcreate(513);
	maus_handle=hhcreate(513);
	XNULLCHECK(rfc_handle);
	XNULLCHECK(maus_handle);

	leer_grp=g;

	if (keep_locked)
		mode="r+";

	sprintf(fname,"%s/%s/etc/groups",opts.spoolbase,action);
	groups_file = fopen (fname, mode);
	if (!groups_file)
	{
		int first_errno=errno;
		sprintf(fname,"%s/groups",opts.spoolbase);
		groups_file = fopen (fname, mode);
		if (!groups_file && first_errno==ENOENT) {
			sprintf(fname,"%s/%s/etc/groups",opts.spoolbase,action);
			groups_file = fopen (fname, "a+");
		}
	}

	if (!groups_file)
	{
		log (LOG_ERR, _("cannot open `groups' %s: %s"), fname, strerror (errno));
		return FALSE;
	}

#ifdef CAN_LOCK_FILES
	if (keep_locked)
		ret=fd_writew_lock(fileno(groups_file),0,SEEK_SET,1);
	else
		ret=fd_readw_lock(fileno(groups_file),0,SEEK_SET,1);

	if (ret != 0) {
		log (LOG_ERR, _("cannot lock %s: %s"), fname, strerror (errno));
		fclose(groups_file);
		groups_file=NULL;
		return FALSE;
	}
#endif

	new=XMALLOC(sizeof(grp_data_t));
	*new=leer_grp;

	while (fgets (str, sizeof (str) - 1, groups_file))
	{
		char *inhalt;
		linenumber++;

		/* 
		 * Zeilenende weg
		 */
		p=strpbrk(str,"\r\n");
		if (p)
			*p='\0';
		if (!*str || !strncmp(str,"---",3))
			continue;
		p=str;
		while (isspace(*p))
			p++;
		if (!*p)
			continue;
		inhalt=strchr(p,'=');
		if (!inhalt)
			continue;

		/* Variablennamen terminieren */

		*inhalt++='\0';
		strtrim(inhalt);
		strtrim(p);

		if (!strcasecmp(p,"group"))
		{
			if (new->rfcname && new->mausname)
			{
				if (!grp_append_new(new))
				{
					fclose(groups_file);
					groups_file=NULL;
					return FALSE;
				}
				new=XMALLOC(sizeof(grp_data_t));
				*new=leer_grp;
			}
			new->rfcname=XSTRDUP(inhalt);
		}
		else if (!strcasecmp(p,"mausname"))
			new->mausname=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"approve") || !strcasecmp(p,"approved"))
			new->approve=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"userinfo") || !strcasecmp(p,"info"))
			new->userinfo=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"followup") || !strcasecmp(p,"followup-to"))
			new->default_followup=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"usenet-distribution") 
			|| !strcasecmp(p,"usenet-dist"))
			new->usenet_distribution=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"mausnet-distribution") 
			|| !strcasecmp(p,"mausnet-dist"))
			new->mausnet_distribution=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"boxen") || !strcasecmp(p,"box"))
		{
#ifdef BOXEN
			new->boxen=XSTRDUP(inhalt);
#endif
		}
		else if (!strcasecmp(p,"path-entry") || !strcasecmp(p,"path"))
			new->path_entry=XSTRDUP(inhalt);
		else if (!strcasecmp(p,"mime"))
		{
			new->mime_headers=TRI_FALSE;
			new->mime_body=TRI_FALSE;
			if (tolower(*inhalt)=='h')
				new->mime_headers=TRUE;
			else if (tolower(*inhalt)=='b')
			{
				new->mime_headers=TRUE;
				new->mime_body=TRUE;
			}
		}
		else if (!strcasecmp(p,"checkgroups"))
		{
			new->checkgroups=parse_bool(inhalt);
		}
		else if (!strcasecmp(p,"itg"))
			new->itg=parse_bool(inhalt);
		else if (!strcasecmp(p,"no-import"))
			new->no_import=parse_bool(inhalt);
		else if (!strcasecmp(p,"no-export"))
			new->no_export=parse_bool(inhalt);
		else
			fprintf(stderr,_("groups: %s: unknown variable\n"),p);
	} /* while */
	if (!keep_locked) {
		fclose(groups_file);
		groups_file=NULL;
	}

	if (new->rfcname && new->mausname)
	{
		if (!grp_append_new(new))
			return FALSE;
	}
	return TRUE;
}

/* Einen neuen Mausnamen in die Liste eintragen */
BOOLEAN
grp_append_new(grp_data_t *new)
{
	char buf[256];
	DYN_ENTRY item;
	DYN_ENTRY *found;
	hash_list_t *hl;

	/* Wenn das Hauptprogramm keine Gruppenliste initialisiert hat, dann 
	 * brauchen wir hoffentlich auch keine */
	if (!maus_handle || !rfc_handle)
		return TRUE;

	canon_grp(new->rfcname,buf);
	item.crc=crc32ansi(buf,strlen(buf));

	hl=XMALLOC(sizeof(hash_list_t));
	hl->data=new;
	found=hhsearch(rfc_handle,&item,DYN_FIND);
	if (found)
	{
		/* da ist schon ein Eintrag fr die Gruppe (das ist bei
		 * mehreren Boxen oder mehreren Benutzern normal!)
		 */
		((hash_list_t *)(found->u.ptr))->next=hl;
		found->u.ptr=hl;
	}
	else
	{
		item.u.ptr=hl;
		if (!hhsearch(rfc_handle,&item,DYN_ENTER))
			error(1,0,_("%s: out of memory"),"grp_append_new");
	}

	canon_grp(new->mausname,buf);
	item.crc=crc32ansi(buf,strlen(buf));
	hl=XMALLOC(sizeof(hash_list_t));
	hl->data=new;
	found=hhsearch(maus_handle,&item,DYN_FIND);
	if (found)
	{
		/* da ist schon ein Eintrag fr die Gruppe (das ist bei
		 * mehreren Boxen oder mehreren Benutzern normal!)
		 */
		((hash_list_t *)(found->u.ptr))->next=hl;
		found->u.ptr=hl;
	}
	else
	{
		item.u.ptr=hl;
		if (!hhsearch(maus_handle,&item,DYN_ENTER))
			error(1,0,_("%s: out of memory"),"grp_append_new");
	}
	
	if (!anker)
		anker=new;
	else
		last->next=new;
	new->next=NULL;
	last=new;
	return TRUE;
}


/* Liefert den mausinternen Namen der mit Rfcnamen gegebenen Gruppe zurck */
grp_data_t *
grp_suche_mausname(const char *mausname, const char *box)
{
	DYN_ENTRY item;
	DYN_ENTRY *found;
	hash_list_t *hl;
	char *buf;

	/* Wenn das Hauptprogramm keine Gruppenliste initialisiert hat, dann 
	 * brauchen wir hoffentlich auch keine */
	if (!maus_handle)
		return NULL;

	buf=alloca(strlen(mausname)*2+1);
	canon_grp(mausname,buf);
	item.crc=crc32ansi(buf,strlen(buf));

	found=hhsearch(maus_handle,&item,DYN_FIND);
	if (!found)
		return NULL;
	hl=found->u.ptr;

#ifdef BOXEN
	/* Ok - die Gruppe ist in der Liste. Jetzt mssen wir nur noch 
	 * feststellen, ob und welcher Eintrag auf die Box pat 
	 */
	if (!box)
		return hl->data;
	
	while (hl)
	{
		grp_data_t *g=hl->data;
		if ((g->boxen[0]=='*' && g->boxen[1]=='\0')
			|| !strcasecmp (box,g->boxen))
			return g;
		hl = hl->next;
	}
	return NULL; /* nichts pat */
#else
	return hl->data;
#endif
}

grp_data_t *
grp_suche_rfcname(const char *rfcname, const char *box)
{
	DYN_ENTRY item;
	DYN_ENTRY *found;
	hash_list_t *hl;
	char *buf;

	/* Wenn das Hauptprogramm keine Gruppenliste initialisiert hat, dann 
	 * brauchen wir hoffentlich auch keine */
	if (!rfc_handle)
		return NULL;

	buf=alloca(strlen(rfcname)*2 +1);
	canon_grp(rfcname,buf);
	item.crc=crc32ansi(buf,strlen(buf));

	found=hhsearch(rfc_handle,&item,DYN_FIND);
	if (!found)
		return NULL;
	hl=found->u.ptr;

#ifdef BOXEN
	/* Ok - die Gruppe ist in der Liste. Jetzt mssen wir nur noch 
	 * feststellen, ob und welcher Eintrag auf die Box pat 
	 */
	if (!box)
		return hl->data;
	
	while (hl)
	{
		grp_data_t *g=hl->data;
		if ((g->boxen[0]=='*' && g->boxen[1]=='\0')
			|| !strcasecmp (box,g->boxen))
			return g;
		hl = hl->next;
	}
	return NULL; /* nichts pat */
#else
	return hl->data;
#endif
}

BOOLEAN
grp_save_list(void)
{
	grp_data_t *akt;

	assert(groups_file);
	rewind(groups_file);
	if (-1==ftruncate(fileno(groups_file),0)) {
		log(LOG_ERR,_("cannot fruncate `groups' file: %s"),strerror(errno));
		fclose(groups_file);
		groups_file=NULL;
		return FALSE;
	}

	akt=anker;
	while (akt) {
		if (!akt->dont_save) {
			fprintf(groups_file,"Group=%s\n",akt->rfcname);
			fprintf(groups_file,"\tMausname=%s\n",akt->mausname);
#ifdef BOXEN
			fprintf(groups_file,"\tBoxen=%s\n",akt->boxen);
#endif
			if (akt->userinfo)
				fprintf(groups_file,"\tUserinfo=%s\n",akt->userinfo);
			if (akt->approve)
				fprintf(groups_file,"\tApprove=%s\n",akt->approve);
			if (akt->default_followup)
				fprintf(groups_file,"\tFollowup=%s\n",akt->default_followup);
			if (akt->checkgroups)
				fprintf(groups_file,"\tCheckgroups=true\n");
			if (akt->itg)
				fprintf(groups_file,"\tITG=true\n");
			if (akt->no_import)
				fprintf(groups_file,"\tno-import=true\n");
			if (akt->no_export)
				fprintf(groups_file,"\tno-export=true\n");
			if (akt->mime_headers==TRI_TRUE)
				fprintf(groups_file,"\tMIME=headers\n");
			else if (akt->mime_headers==TRI_FALSE)
				fprintf(groups_file,"\tMIME=none\n");
			if (akt->path_entry!=NULL)
				fprintf(groups_file,"\tPath-Entry=%s\n",akt->path_entry);
			if (akt->usenet_distribution!=0)
				fprintf(groups_file,"\tUsenet-Distribution=%s\n",
					akt->usenet_distribution);
			if (akt->mausnet_distribution!=0)
				fprintf(groups_file,"\tMausnet-Distribution=%s\n",
					akt->mausnet_distribution);
			fprintf(groups_file,"--------------------------------------\n");
		}
		akt=akt->next;
	}
	fclose(groups_file);
	groups_file=NULL;
	return TRUE;
}


/* Gruppennamenskonvertierung */

/* Was soll passieren, falls ein Namensteil zu lang (>14 Zeichen) ist?
 * #undef SKIP_TO_DOT  -> Punkt zwischenschieben.
 * #define SKIP_TO_DOT -> alles bis zum nchsten Punkt ignorieren.
 * Letzteres mte ausreichen, um eindeutige Namen zu erzeugen, erzeugt
 * aber eindeutig die schneren Namen.
 */
#define SKIP_TO_DOT

/*
 * Liste bekannter Gruppenhierarchienamen, fr die kein "maus." als
 * Prefix vorgehngt wird.
 */
const char *known_prefix[] =
{
	"alt",
	"aus",
	"bionet",
	"bit",
	"biz",
	"bln",
	"clari",
	"comp",
	"de",
	"fido",
	"fj",
	"gnu",
	"k12",
	"maus",
	"misc",
	"news",
	"rec",
	"schule",
	"sci",
	"soc",
	"talk",
	"vmsnet",
	"zer",
	NULL
};

#define IS_SPECIAL(chr) (strchr("+-._",chr))
BOOLEAN 
grp_create_rfcname (unsigned char *unconverted, unsigned char *to)
{
	unsigned char *p;
	unsigned char *q;
	unsigned char *last_dot;
	unsigned char c;
	unsigned char origc;
	unsigned char *buf;
	const char **prefix;

#ifdef SKIP_TO_DOT
	int skip_to_dot = 0;
#endif

	buf = entumlauten1 (unconverted);
	XNULLCHECK(buf);

	/* Feststellen, ob ein "maus."-Prefix vor den Namen gestellt werden
	 * mu.
	 */
	*to = '\0';
	prefix = known_prefix;
	while (*prefix)
	{
		size_t l;
		l=strlen(*prefix);
		if (!strncasecmp (buf, *prefix, l))
		{
			if (buf[ l ] == '.')
				break;
		}
		prefix++;
	}
	if (!*prefix)
	{
		/* notorischen Fall abfangen */
		if (!strcmp(buf,"maus"))
			sprintf (to, "maus.maus");
		else
			sprintf (to, "%s.%s.", opts.gprefix, opts.pollsite);
	}

	q = to + strlen (to);
	last_dot = q - 1;

	for (p = buf; *p; p++)
	{
		unsigned char last_char;
		if (p==buf)	
			last_char='\0';
		else 
			last_char=p[-1];

		/* 
		 * Zeichenmenge auf Kleinbuchstaben und Ziffern sowie "+-_."
		 * beschrnken.
		 */
		origc = *p;
		if (isupper (origc))
			c = tolower (origc);
		else if (origc=='&' || origc=='/')
			c = '+';
		else if (islower (origc) || isdigit (origc) || origc == '.'
			|| origc == '+' || origc == '-' || origc == '_' || origc == '=')
			c = origc;
		else
		{
			/* alles andere in Punkte verwandeln */
			c = '.';
		}

		/* 
		 * Das Anfangszeichen der Komponente prfen 
		 */
		if (q - last_dot == 1)
		{
			/* "subsequent components SHOULD begin with a letter" */
			if (!(islower (c)))
				continue;
		}
		if (q - last_dot == 14)
		{
			/* 14 Zeichen ist die Maximallnge fr Namensteile */
#ifndef SKIP_TO_DOT
			last_dot = q;
			*q++ = '.';
			last_char = '.';
#else
			skip_to_dot = 1;
#endif
		}

		/* Den Punkt behandeln */
		if (c == '.')
		{	
#ifdef SKIP_TO_DOT
			skip_to_dot = 0;
#endif
			/* "+" wg. c++ besondert behandeln */
			if (IS_SPECIAL(last_char) && last_char!='+')
			{
				q[-1]='.';
				last_dot=q-1;
			}
			else
			{
				*q++='.';
				last_dot=q;
			}
			continue;
		}
		/* 
		 * Auseinanderfolgende Sonderzeichen vermeiden, mit Ausnahme 
		 * von "++" wie in "c++" 
		 */
		if (IS_SPECIAL(c) && IS_SPECIAL(last_char) && c!='+' && last_char!='+')
			continue;
#ifdef SKIP_TO_DOT
		if (skip_to_dot)
				continue;
#endif
		*q++ = c;
	}
	while (q > to && q[-1] == '.')
		*(--q) = '\0';
	*q = '\0';
	free (buf);
	return TRUE;
}

#if 0

Alte Version

static void
convert_groupname (unsigned char *unconverted, unsigned char *to)
{
	unsigned char *p;
	unsigned char *q;

	if ((p = get_rfcname (opts.pollsite, unconverted)) != NULL)
	{
		strcpy (to, p);
	}
	else
	{
		sprintf (to, "%s.%s.", opts.gprefix, opts.pollsite);
		p = unconverted;
		q = to + strlen (to);

		while (*p)
		{
			if (isupper (*p))
				*q++ = tolower (*p);
			else if (islower (*p) || *p == '.')
				*q++ = *p;
			else
				*q++ = '_';
			p++;
		}
		put_mausname (unconverted, opts.pollsite, to, NULL);
	}
}
#endif
