/* Mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 1996 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   maus2rfc: Wandeln von MausTausch nach RFC.
*/

#include "config.h"

#include <sys/types.h>
#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <syslog.h>
#include <unistd.h>
#include <getopt.h>
#include <assert.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#include "maus.h"
#include "srfcmail.h"
#include "srfcnews.h"
#include "groups.h"
#include "domains.h"
#include "long-options.h"
#include "msg_io.h"

#ifdef WANT_NNTP
#	include "nntp.h"
#endif

#define HEADLEN	1024
enum messagetyp { NO_MSG, MAIL_MSG, NEWS_MSG, IGNORE_MSG };
enum specialtyp { NO_SPECIAL, INFO_SPECIAL, LOG_SPECIAL };
enum messagetyp this_message;
enum specialtyp this_special;

/* ------------------------------------------------------------------------- */
metamsg_t meta;

int work_only;
int dont_use_nntp;
int dont_use_smtp;

static BOOLEAN maus2rfc (const char *fname);
static int count_refs(char *look);
static int prepend_refs(int ab, char *look, metamsg_t *msg);

static int argument_only_convert=0;
static FILE *new_status_f=NULL;


/* ------------------------------------------------------------------------- */

static void parse_cmdlog(buf_t *neu, char *start, char **next)
{
	const char *rueckmeldungen="?!";
	BOOLEAN err=FALSE;
	char *dies=start;

	/* wir parsen alle Zeilen nach dem Kommando und sehen nach,
	 * ob da eine Fehlermeldung kommt. Das fngt auch den
	 * eher unwahrscheinlichen Fall
	 * :"Kommando"
	 * :!bemerkung
	 * :?Fehler
	 * ab.
	 */
	do
	{
		dies=strpbrk(dies,"\r\n");
		if (!dies)
			break;
		dies+=2; /* \r\n ueberspringen */
	} while (strchr(rueckmeldungen,*dies));
	*next=dies;
	if (!err || !opts.smart_LOG || !neu)
		return;
	dies=start;
	do
	{
		char c;
		char *line=dies;
		dies=strpbrk(dies,"\r\n");
		if (!dies)
			break;
		c=*dies;
		*dies=0;
		ADD_TO_BUF(neu,line);
		ADD_TO_BUF(neu,"\n");
		*dies=c;
		if (c=='\r')
			dies++;
		if (*dies=='\n')
			dies++;
	} while (strchr(rueckmeldungen,*dies));
}

static void parse_msglog(buf_t *neu, char *start, char **next)
{
	/* Das Format des Logfiles:
	 *  :#12345
	 *  :!unwichtiges Gelaber
	 *  :?Haben sie eine Zeitmaschine (Datum/Uhrzeit in der Zukunft)
	 *  :=A123@...
	 * oder
	 *  :#12345
	 *  :?Das duerfen sie nicht.
	 * Ob eine Nachricht akzeptiert wurde, kann man also bei der ersten
	 * Gleichheitszeichenzeile sagen.
	 */
	char *dies;
	const char *rueckmeldungen="?!=";
	char *nullp;
	char nullc='\0'; /* keep gcc happy */
	char fname[PATH_MAX];
	char *mausid=NULL;
	char *transferid=NULL;
	char *errtext=NULL;
	char *spruchtext=NULL;
	int ok;
	BOOLEAN err=FALSE;

	dies=start;

	ok=0; /* noch keine "="-Zeile bekommen. */

	do
	{
		nullp=strpbrk(dies,"\r\n");
		if (!nullp)
		{
			*next=dies+strlen(dies);
		}
		else
		{
			char *p;
			p=nullp+1;
			while (*p=='\r' || *p=='\n')
				p++;
			*next=p;
		}

		/* start+1 zeigt jetzt auf die ID im Infile, 
		 * next auf die nchste Zeile, dies auf diese Zeile 
		 */
		if (nullp)
		{
			nullc=*nullp;
			*nullp='\0';
		}
		switch(*dies)
		{
		case '=':
			mausid=XSTRDUP(dies+1);
			ok=1;
			break;
		case '?': /* ein Fehler */
			err=TRUE;
		case '!': /* irgendein Spruch. */
			{
				char **indirect=(*dies=='!') ? &spruchtext : &errtext;
				if (!*indirect)
					*indirect=concat("\t",dies+1,"\r\n",NULL);
				else
				{
					*indirect=concat(*indirect,"\t",dies+1,"\r\n",NULL);
					XNULLCHECK(*indirect);
				}
			}
			break;
		case '#':
			transferid=XSTRDUP(dies+1);
			break;
		}
		if (nullp)
			*nullp=nullc;
		dies=*next;

	} while (*dies && strchr(rueckmeldungen,*dies));

	if (err && opts.smart_LOG && neu)
	{
		dies=start;
		do
		{
			char c;
			char *line=dies;
			dies=strpbrk(dies,"\r\n");
			if (!dies)
				break;
			c=*dies;
			*dies=0;
			ADD_TO_BUF(neu,line);
			ADD_TO_BUF(neu,"\n");
			*dies=c;
			if (c=='\r')
				dies++;
			if (*dies=='\n')
				dies++;
		} while (strchr(rueckmeldungen,*dies));
	}

	assert(transferid);

	if (opts.boxtyp==BOXTYP_MAUS && !strncmp(transferid,"MAIL",4))
	{
		/* 
		 *	Die MAUS <7.94 wandelte die Benutzerids in Uppercase ...
		 */
		char *p=transferid;
		while (*p)
		{
			*p=tolower(*p);
			p++;
		}
	}

	/* Statusmeldungen ... */
	if (strchr(transferid,'@') && !mausid)
	{
		if (errtext)
		{
			char *str;
			char *p,*q;
			/* eine Fassung der Fehlermeldung ohne \r\n erzeugen */
			str=XMALLOC(strlen(errtext));
			for (p=errtext,q=str;*p;p++)
			{
				if (*p=='\t')
					*q++=' '; /* space */
				else if (*p!='\r')
					*q++=*p;
			}
			*q=0;
			msglog ("! %s Error (statusmsg) : %s", transferid,str);
			free(str);
		}
		if (spruchtext)
		{
			char *str,*p,*q;
			/* eine Fassung der Fehlermeldung ohne \r\n erzeugen */
			str=XMALLOC(strlen(spruchtext));
			for (p=spruchtext,q=str;*p;p++)
			{
				if (*p=='\t')
					*q++=' '; /* space */
				else if (*p!='\r')
					*q++=*p;
			}
			*q=0;
			msglog ("! %s Notice (statusmsg) : %s", transferid,str);
			free(str);
		}
		if (!spruchtext && !errtext)
			msglog ("! %s (statusmsg) ok", transferid);
		free (spruchtext);
		free (errtext);
		free (transferid);
		return;
	}

	if (!mausid && !errtext)
	{
		free (spruchtext);
		free (transferid);
		return;
	}

	sprintf (fname, "%s/%s/outgoing/%s", 
		opts.spoolbase, action, transferid);

	if (opts.workmode==GATE)
	{
		/* dirty hack fr Gatemodus: Da die Transfer-ID des Batchfiles
		 * als Message-ID mitgeschickt wurde, bekommen wir nun eine
		 * Fehlermeldung fr das Batchfile, und damit die Mglichkeit,
		 * es zu lschen.
		 */
		if (!strchr(transferid,'@'))
		{
			free (spruchtext);
			free (transferid);
			free (mausid);
			free (errtext);
			unlink(fname);
			return;
		}
	}

	if (errtext && !mausid)
	{
		char cmd[PATH_MAX*3];
		char *str;
		char *q;
		char *p;
		int was_mail=0;

		/* Struktur mit den Daten der gescheiterten Nachricht */
		metamsg_t done;
		/* Alles schoen als leer markieren */
		memset(&done,0,sizeof(metamsg_t));

		/* eine Fassung der Fehlermeldung ohne \r\n erzeugen */
		str=XMALLOC(strlen(errtext));
		for (p=errtext,q=str;*p;p++)
		{
			if (*p=='\t')
				*q++=' '; /* space */
			else if (*p!='\r')
				*q++=*p;
		}
		*q=0;

		/* Einmal per Syslog */
		if (opts.workmode==USER)
			log (LOG_WARNING, "Message %s failed: %s", fname, str);
		/* und ins Logfile */
		msglog ("! [%s]: FAILED: %s", transferid,str);
		free(str);

		/* So, jetzt die Message bouncen. Aber vorsichtig: Im Gateway-
		 * betrieb darf das nur fr Mail geschehen, im Userbetrieb
		 * sollte es fr News und Mail passieren.
		 */
		was_mail=!strncmp(transferid,"mail",4);
		if (was_mail || opts.workmode==USER)
		{
			/* Struktur mit der Bouncemeldung. */
			static long id_counter=0;
			metamsg_t msg;
			char zwi[1024];
			char idzwi[256];
			char ozwi[2048];
			char *absender=NULL;
			
			memset(&msg,0,sizeof(metamsg_t));
			msg.is_mail=was_mail;

			sprintf(zwi,"Mauscon@%s",opts.mysitedomain);
			msg.env_from=XSTRDUP(zwi);

			msg.head_from=XMALLOC(sizeof(metalist_t));
			msg.head_from->next=NULL;
			msg.head_from->ext=NULL;
			msg.head_from->text=XSTRDUP(zwi);

			/* this is ok */
			msg.subject=XSTRDUP("Message failed");

			{
				time_t t=time(NULL);
				msg.tm_date=XMALLOC(sizeof(struct tm));
				memcpy(msg.tm_date,localtime(&t),sizeof(struct tm));
			}

			sprintf(idzwi,"mauscon-bounce.%ld.%ld@%s",(long) getpid(),
				++id_counter,opts.mysitedomain);
			msg.message_id=XSTRDUP(idzwi);

			sprintf(ozwi,
				"Your message was not accepted by %s, because:\r\n",
					opts.pollsitedomain);
			meta_add_to_body(&msg,ozwi);
			meta_add_to_body(&msg,errtext);
			if (spruchtext)
			{
				sprintf(ozwi,
					"\r\n%s stated the following:\r\n",
						opts.pollsitedomain);
				meta_add_to_body(&msg,ozwi);
				meta_add_to_body(&msg,spruchtext);
			}

			/* Jetzt den Body der Message anhngenA
			 * im Hchstfall 16Kb, mehr lohnt irgendwie nicht.
			 */
			{
				buf_t bouncebuf={NULL,NULL,0};
				if (!rfc_read_message(fname,&bouncebuf) && bouncebuf.start)
				{
					if (strlen(bouncebuf.start)<16*1024)
					{
						meta_add_to_body(&msg,"Failed message follows:\n\n");
						meta_add_to_body(&msg,bouncebuf.start);
					}
					else
					{
						meta_add_to_body(&msg,
							"Failed message excedes 16K - not copied!\n");
					}
					if (was_mail)
						absender=rfc_find_header (bouncebuf.start, 
							"x-mausconnect-env-from");
					if (!absender)
						absender=rfc_find_header (bouncebuf.start, 
							"reply-to");
					if (!absender)
						absender=rfc_find_header (bouncebuf.start, 
							"sender");
					if (!absender)
						absender=rfc_find_header (bouncebuf.start, 
							"from");
				}
				else
				{
					log(LOG_WARNING, _("cannot read %s (in bounce)"),fname);
					meta_add_to_body(&msg,"Sorry - cannot read message!");
				}
				free(bouncebuf.start);
			}
			if (absender)
			{
				msg.env_to=XSTRDUP(absender);
				msg.head_to=XMALLOC(sizeof(metalist_t));
				msg.head_to->next=NULL;
				msg.head_to->ext=NULL;
				msg.head_to->text=XSTRDUP(absender);
				srfcmail(argument_only_convert,&msg);
				meta_free_list(&msg.head_to);

				free(msg.subject);
				free(msg.env_to);
				msg.subject=XSTRDUP("Message failed, gatemasters copy");
			}
			else
			{
				free(msg.subject);
				msg.subject=XSTRDUP("Message failed, sender unknown, waking gatemaster");
			}
			msg.head_to=XMALLOC(sizeof(metalist_t));
			msg.head_to->next=NULL;
			if (absender)
				msg.head_to->ext=NULL;
			else
				msg.head_to->ext=XSTRDUP("Mauscon needs your help");
			msg.head_to->text=XSTRDUP(opts.gatemaster);
			msg.env_to=XSTRDUP(opts.gatemaster);
			srfcmail(argument_only_convert,&msg);
			meta_clean_message(&msg);
			free(absender);
			/* Bounce abgeschlossen */
		} /* soll gebounced werden? */
		sprintf (cmd, "%s/%s/failed/%s", opts.spoolbase, action, transferid);
		if (rename(fname,cmd))
		{
			sprintf(cmd,"mv %s %s/%s/failed/",
				fname,opts.spoolbase,action);
			system(cmd);
		}
		free (spruchtext);
		free (transferid);
		free (mausid);
		free (errtext);
		return;
	}
	/* Nachricht war OK.
	 * Im Usermodus nun die ursprngliche Nachricht canceln 
	 */
	msglog ("  [%s]: sent to %s", transferid,opts.pollsite);
	if (opts.workmode==USER 
		&& *transferid=='m' && transferid[1]=='t' && isdigit(transferid[2])
		&& (work_only==ONLY_ALL || work_only==ONLY_NEWS))
	{
		buf_t buf={NULL,NULL,0};
		char fn2[PATH_MAX];
		sprintf (fn2, "%s/%s/outgoing/%s", 
			opts.spoolbase, action, transferid+2);
		if (0==rfc_read_message(fn2,&buf))
		{
			char *h;
			char *id;
			char *ng;
			id=rfc_find_header(buf.start,"Message-ID");
			ng=rfc_find_header(buf.start,"Newsgroups");
			h=rfc_find_header(buf.start,"Sender");
			if (!h)
				h=rfc_find_header(buf.start,"From");
			if (ng && id && h && 
				!rfc_find_header(buf.start,"X-MausConnect-Env-To"))
			{
				metamsg_t msg;
				
				memset(&msg,0,sizeof(metamsg_t));
				if (!rfc_to_meta(&buf,&msg))
				{
					char *zwi;
					char *end;

					/* Control-Header */

					zwi=XMALLOC(strlen(msg.message_id)
						+sizeof("cancel <>")+1);
					sprintf(zwi,"cancel <%s>",msg.message_id);
					if (msg.control)
						free(msg.control);
					msg.control=zwi;
					msg.transfer_id=XSTRDUP(transferid);

					/* Subject */

					zwi=XMALLOC(strlen(msg.message_id)
						+sizeof("cmsg cancel <>")+1);
					sprintf(zwi,"cmsg cancel <%s>",msg.message_id);
					if (msg.subject)
						free(msg.subject);
					msg.subject=zwi;

					/* Neue Message-ID jetzt erst vergeben */

					zwi=XMALLOC(strlen(msg.message_id)
						+sizeof("cancel-")+1);
					sprintf(zwi,"cancel-%s",msg.message_id);
					free(msg.message_id);
					msg.message_id=zwi;

					free(msg.body.start);
					msg.body.size=1024;
					msg.body.start=XMALLOC(msg.body.size);
					end=stpcpy(msg.body.start,
							   _("original canceled by Mauscon"));
					end=stpcpy(end,"\r\n");
					msg.body.next=end;

					if (msg.distribution)
					{
						free(msg.distribution);
						msg.distribution=NULL;
					}
					if (opts.usenet_dist_lokal)
					{
						msg.distribution=XSTRDUP(opts.usenet_dist_lokal);
						srfcnews(argument_only_convert,&msg);
					}
					meta_clean_message(&msg);
				}
			}
		}
		free(buf.start);
	}
	/* Originaldatei zu news jetzt in jedem Fall lschen, auch wenn 
	 * cancel_original nicht gewnscht ist. Grund: es knnte man
	 * gewnscht gewesen sein - dann liegt es jetzt hier herum ... 
	 */
	if (opts.workmode==USER 
		&& *transferid=='m' && transferid[1]=='t' && isdigit(transferid[2])
		&& (work_only==ONLY_ALL || work_only==ONLY_NEWS)
		&& !opts.cancel_original)
	{
		char fn2[PATH_MAX];
		sprintf (fn2, "%s/%s/outgoing/%s", 
			opts.spoolbase, action, transferid+2);
		unlink(fn2);
	}

	{
		unsigned char ov_fname[PATH_MAX*3];
		char *id=NULL;
		char *to=NULL;
		char *subject=NULL;
		FILE *df;
		df=fopen(fname,"r");
		if (df)
		{
			char buf[1024];
			while (fgets(buf,sizeof(buf),df))
			{
				char *p=strpbrk(buf,"\r\n");
				if (p) *p=0;
				if (!*buf)
					break;
#define MAGIC "message-id:"
				if (!strnhalfcasecmp(buf,MAGIC,sizeof(MAGIC)-1))
				{
					p=buf+sizeof(MAGIC)-1;
					while (isspace(*p))
						p++;
					if (*p)
						id=XSTRDUP(p);
				}
#undef MAGIC
#define MAGIC "subject:"
				if (!strnhalfcasecmp(buf,MAGIC,sizeof(MAGIC)-1))
				{
					p=buf+sizeof(MAGIC)-1;
					while (isspace(*p))
						p++;
					if (*p)
						subject=XSTRDUP(p);
				}
#undef MAGIC
#define MAGIC "x-mausconnect-env-to:"
				if (!strnhalfcasecmp(buf,MAGIC,sizeof(MAGIC)-1))
				{
					p=buf+sizeof(MAGIC)-1;
					while (isspace(*p))
						p++;
					if (*p)
						to=XSTRDUP(p);
				}
#undef MAGIC
			}
			fclose(df);
			if (id && to)
			{
				sprintf(ov_fname,"%s/%s/status/overview",
					opts.spoolbase,action);
				df=fopen(ov_fname,"a");
				if (df)
				{
					fprintf(df,"%s %s\n",mausid,transferid);
					fclose(df);
					sprintf(ov_fname,"%s/%s/status/%s.status",
						opts.spoolbase,action,transferid);
					df=fopen(ov_fname,"w");
					if (df)
					{	
						time_t t=time(NULL);
						fprintf(df,"Message-ID: %s\n",id);
						fprintf(df,"To: %s\n",to);
						if (subject)
							fprintf(df,_("Subject: %s\n"),subject);
						/* End of Header */
						fprintf(df,"\n");
						fprintf(df,_("Versendet: %s\n"),ctime(&t));
						fclose(df);
					}
				}
			}
		}
		free(to);
		free(id);
		free(subject);
	}
	if (opts.keep_sent)
	{
		unsigned char ov_fname[PATH_MAX*3];
		sprintf (ov_fname, "%s/%s/sent2maus/%s", opts.spoolbase, action, transferid);
		if (rename(fname,ov_fname))
		{
			sprintf (ov_fname, "mv %s %s/%s/sent2maus/%s", 
				fname, opts.spoolbase, action, transferid);
			if (system(ov_fname))
				unlink(fname);
		}
	}
	else
		unlink(fname);
	free (spruchtext);
	free (transferid);
	free (mausid);
	free (errtext);
}

static void
work_log (buf_t *bodybuf)
{
	char *this;
	char *next;
	char *p;
	char *nullp;
	char nullc='\0'; /* keep gcc happy */
	unsigned char *body;
	buf_t neu={NULL,NULL,0};
	char *last_marker=NULL;
	BOOLEAN in_meldung=FALSE;

	body=bodybuf->start;


	this=next=body;

	while (this && *this)
	{
		nullp=strpbrk(this,"\r\n");
		if (nullp)
		{
			p=nullp+1;
			while (*p=='\r' || *p=='\n')
				p++;
			next=p;
		}
		else
			next=strlen(this)+this;
		switch(*this)
		{
		case '\"':
			last_marker=this;
			in_meldung=FALSE;
			parse_cmdlog(&neu,this,&next);
			break;
		case '#':
			last_marker=this;
			in_meldung=FALSE;
			parse_msglog(&neu,this,&next);
			break;
		case '$': /* Infofilequittungen */
			in_meldung=FALSE;
			last_marker=this;
			p=strchr(this,'=');
			if (p)
			{
				*p='\0';
				if (nullp)
				{
					nullc=*nullp;
					*nullp=0;
				}
				info_new_crc(this+1,strtol(p+1,NULL,10));
				if (nullp)
				{
					*nullp=nullc;
				}
				*p='=';
			}
			break;
		default:
			break;
		} /* switch */
		this=next;
	}

	if (opts.smart_LOG) {
		free(bodybuf->start);
		*bodybuf=neu;
	}
}

static int 
count_refs(char *look)
{
	char *p;
	int  count=0;
	while (*look)
	{
		while (*look && *look!='<')
			look++;
		p=strchr(look,'>');
		if (!p)
			break;

		count++;
		look=p+1;
		while (isspace(*look) || *look==',')
			look++;
	}
	return count;
}

static int 
prepend_refs(int ab, char *look, metamsg_t *msg)
{
	metalist_t *last=NULL;
	char *p;
	int  count=0;
	while (*look)
	{
		while (*look && *look!='<')
			look++;
		p=strchr(look,'>');
		if (!p)
			break;

		if (count>=ab || count==0)
		{
			metalist_t *new=NULL;	
			*p='\0';
			meta_add_list(&new,look+1,NULL,1);
			if (new)
			{
				if (!last)
				{
					new->next=msg->references;
					msg->references=new;
				}
				else
				{
					new->next=last->next;
					last->next=new;
				}
				last=new;
			}
			*p='>';
		}
		count++;
		look=p+1;
		while (isspace(*look) || *look==',')
			look++;
	}
	return count;
}

static const char *
status_name(const char st)
{
	switch(st)
	{
	case 'N': return _("not read"); break;
	case 'Z': return _("deffered"); break;
	case 'B': return _("answered"); break;
	case 'G': return _("read"); break;
	case 'W': return _("forwarded"); break;
	case 'M': return _("in MausNet"); break;
	case 'A': return _("arrived at destination bbs"); break;
	case 'Y': return _("arrived at gateway"); break;
	case 'K': return _("copied"); break;
	case 'T': return _("exported to user"); break;
	default: return _("unknown status"); break;
	}
}

static void
send_new_status(void)
{
	char fname[PATH_MAX];
	metamsg_t msg;
	buf_t buf={NULL,NULL,0};

	fclose(new_status_f);
	sprintf(fname,"%s/%s/status/status.new",
		opts.spoolbase,action);

	if (rfc_read_message(fname,&buf))
		return; /* ??? was ist das? */
	
	memset(&msg,0,sizeof(metamsg_t));
	if (rfc_to_meta(&buf,&msg))
		return;
	msg.is_mail=1;
	if (srfcmail(FALSE,&msg))
		unlink(fname);
	else 
		log(LOG_NOTICE,_("cannot send mailstatuschange to gatemaster"));
	free(buf.start);
}

static void
work_incoming_status(metamsg_t *msg)
{
	char fname[PATH_MAX];
	char buf[1024];
	char *t_id=NULL;
	FILE *of;
	time_t t;
	struct tm *tm;

	sprintf(fname,"%s/%s/status/overview",opts.spoolbase,action);
	of=fopen(fname,"r");
	if (!of)
		return;

	while (fgets(buf,sizeof(buf),of))
	{
		char *p=strpbrk(buf,"\r\n");
		if (p) *p=0;
		p=strpbrk(buf,"\t ");
		if (!p)
			continue;
		*p++=0;
		/* yessir, i'm paranoid */
		while (isspace(*p))
			p++;
		if (!*p)
			continue;
		if (!strcasecmp(buf,msg->message_id))
		{
			t_id=XSTRDUP(p);
			break;
		}
	}
	fclose(of);

	if (!t_id)
		return;

	sprintf(fname,"%s/%s/status/%s.status",
		opts.spoolbase,action,t_id);
	free(t_id);

	of=fopen(fname,"a+");
	if (!of)
	{
		free(t_id);
		return;
	}
	t=time(NULL);
	tm=localtime(&t);
	strftime(buf,sizeof(buf),
		_("status change received %d.%m.%y, %H:%M"),tm);
	fprintf(of,"%s\n",buf);
	if (msg->maus_status[1])
	{
		mtzeit_to_time_t(msg->maus_status+1,&t);
		tm=localtime(&t);
		strftime(buf,sizeof(buf),_("status changed %d.%m.%y, %H:%M"),tm);
		fprintf(of,"%s\n",buf);
	}

	fprintf(of,_("new state; %s (%c)"),
		status_name(msg->maus_status[0]),msg->maus_status[0]);
	fputs("\n\n\n",of);

	if (opts.mail_status_change)
	{
		rewind(of); /* von nun an lesen wir dort */


		if (!new_status_f)
		{
			char *tmp;
			sprintf(fname,"%s/%s/status/status.new",
				opts.spoolbase,action);
			new_status_f=fopen(fname,"a");
			if (!new_status_f)
			{
				fclose(of);
				return;
			}

			t=time(NULL);
			tm=localtime(&t);
			strftime(buf,sizeof(buf),"%d.%m.%y, %H:%M Uhr",tm);
			tmp=concat("Mauscon","@",opts.mysitedomain,NULL);
			XNULLCHECK(tmp);
			fprintf(new_status_f,"X-Mausconnect-Env-To: <%s>\n",opts.gatemaster);
			fprintf(new_status_f,"X-Mausconnect-Env-From: <%s>\n",tmp);
			fprintf(new_status_f,"From: <%s>\n",tmp);
			fprintf(new_status_f,"To: <%s>\n",opts.gatemaster);
			fprintf(new_status_f,"Message-Id: <NeueStatusMsgs.%lu@%s>\n",
				(unsigned long) time(NULL), opts.mysitedomain);

			free(tmp);
			fprintf(new_status_f,
				_("%s: status changes received at %s: %s"),
					"Subject", buf, "\n\n");
		}

		while (fgets(buf,sizeof(buf),of))
		{
			char *p=strpbrk(buf,"\r\n");
			if (p) *p=0;

			fprintf(new_status_f,"%s\n",buf);
			if (!*buf)
				break;
		}

		if (msg->maus_status[1])
		{
			mtzeit_to_time_t(msg->maus_status+1,&t);
			tm=localtime(&t);
			strftime(buf,sizeof(buf),
					 _("status changed %d.%m.%y, %H:%M"),tm);
			fprintf(new_status_f,"%s\n",buf);
		}

		fprintf(new_status_f,
				_("new state: %s (%c)%s"),
				status_name(msg->maus_status[0]),msg->maus_status[0],
				"\n\n\n");
	}
	fclose(of);

}

BOOLEAN
maus2rfc (const char *inputfilename)
{
	mio_t *chan;
	char errbuf[PATH_MAX];
	mio_retcode_t mr;

	mr=mio_maustausch.open(inputfilename,"r",&chan,errbuf);
	if (mr!=MIO_OK)
	{
		log(LOG_ERR,_("cannot open %s: %s"), inputfilename, errbuf);
		return FALSE;
	}

	do
	{
		mr=chan->fn->import(chan,&meta);
		if (mr==MIO_EOF)
			break;

		if (mr==MIO_RETRY)
			continue; /* Log done */

		if (mr==MIO_ERROR)
		{
			log(LOG_CRIT,chan->errtext ? chan->errtext : _("unknown error"));
			chan->fn->close(chan);
			return FALSE;
		}

		if (meta.subject && !strncmp(meta.subject,"cmsg ",5)
			&& !meta.also_control && !meta.control)
		{
			/* das passiert z.B., wenn eine Control-Message in eine
			 * Sysopgruppe weitergeleitet wird.
			 * Und das Newssystem darf das dann nicht als control-
			 * message verstehen.
			 * Yep, this is a hack ...
			 */
			meta.subject[4]='_';
		}
#ifdef WANT_NNTP
		if ((meta.env_to || meta.newsgroups) && !meta.maus_special_id
			&& meta.message_id && meta.references
			&& opts.max_references>1)
		{
			char *ref;
			/* 
			 * Versuchen, die References zu verlngern, wenn nur 
			 * eine vorliegt.
			 */
			if (!meta.references->next)
				ref=nntp_inquire_get_headerentry(meta.references->text,"references");
			else 
				ref=NULL;

			if (ref)
			{
				int count;

				count=count_refs(ref);
				if (count)
				{
					int ab;
					ab=count-opts.max_references+2;
					if (ab<0)
						ab=0;
					prepend_refs(ab, ref, &meta);
				}
			}
		}
#endif
		if (meta.maus_special_id)
		{
			minfo_t *found;
			if (work_only==ONLY_ALL || work_only==ONLY_SPECIAL)
			{

				found=info_find(meta.maus_special_id);

				if (found)
					info_work_new_infofile(&meta,found);
					/* Alles auer HEAD und REN mailen */
				else if (strcasecmp(meta.maus_special_id,"HEAD")
					&& strcasecmp(meta.maus_special_id,"REN")
					&& meta.body.start && *meta.body.start)
				{
					if (!strcasecmp(meta.maus_special_id,"LOG"))
						work_log(&meta.body);
					if (meta.body.start)
						meta.env_to=XSTRDUP(opts.mailuserdomain);
				}
			}
		}
		/* NICHT else if (meta.env_to), da das zwischenzeitlich
		 * gesetzt worden sein knnte.
		 */
		if (meta.env_to)
		{
			/* mail */
			if (work_only==ONLY_ALL || work_only==ONLY_MAIL
				|| (work_only==ONLY_SPECIAL && meta.maus_special_id))
				srfcmail(argument_only_convert,&meta);

			if (opts.autostatus && !meta.maus_special_id)
			{
				char fname[PATH_MAX];
				char lockfname[PATH_MAX];
				int fatal_err=0;
				FILE *status_f;

				sprintf(lockfname,"%s/%s/etc/LOCK",opts.spoolbase,action);
				while (lock_create (lockfname, &fatal_err))
				{
					if (fatal_err)
					{
						log(LOG_ERR,_("cannot lock %s: %s"),lockfname,
							strerror(errno));
						break;
					}
					sleep(5);
				}
				if (!fatal_err)
				{
					sprintf(fname,"%s/%s/etc/status.msg",opts.spoolbase,action);
					status_f=fopen(fname,"a");
					if (status_f)
					{
						char z[30];
						time_t t;
						fprintf(status_f,"#%s\n",meta.message_id);
						fprintf(status_f,"B%c\n",opts.autostatus);
						t=time(NULL);
						time_t_to_mtzeit(&t,z);
						fprintf(status_f,"E%s\n",z);
						fclose(status_f);
					}
					lock_delete(lockfname);
				}
			}
		}
		else if (meta.maus_status)
		{
			work_incoming_status(&meta);
		}
		else if (meta.newsgroups)
		{
			if (work_only==ONLY_ALL || work_only==ONLY_NEWS)
				srfcnews(argument_only_convert,&meta);
		}
	} while (mr==MIO_RETRY || mr==MIO_OK);
	chan->fn->close(chan);
	return TRUE;
}

static void
usage(int exitcode, const char *what)
{
    FILE *f=stdout;

    if (exitcode)
    {
        if (what)
            fprintf(stderr, "%s: %s\n",program_name,what);
        fprintf (stderr, _("Try `%s --help' for more information.\n"),
            program_name);
        exit(exitcode);
    }

    fprintf(f, "%s - %s\n", program_name, PACKAGE_VERSION);

    fprintf(f,_("Usage: %s [options] file ...\n"), program_name);
    fputs(_("Convert and send MausTausch to RFC\n"),f);
	    fputs(_(
"  -c  --cnf-file FILE         use config file FILE\n"
"      --help                  this text\n"
"  -i, --input FILE            read input (outfile) from FILE\n"
"      --dont-use-nntp         do not use NNTP, just spool into incoming\n"
"      --dont-use-smtp         do not use SMTP, just spool into incoming\n"
"      --only-mail             send only mail, no news, no spezial files\n"
"      --only-news             send only news, no mail, no spezial files\n"
"      --only-special          send only special files, no news, no mail\n"
"  -O, --only-convert          do not send anything, just print to stdout\n"
"  -v  --version               show version info\n"
"\n"
"short options use the same arguments as the long ones\n"
    ),f);
    exit(exitcode);
}

static void
usage1(int ecode)
{
	usage(ecode,NULL);
}

int
main (int argc, char **argv)
{
	/* Outfile.txt for Reading */
	const char *inputfilename="-";
	const char *cnffname=SYSCONFDIR "/" CNFNAME;
	int exitcode=0;

	struct option cmdlinopts[] =
	{
		{"cnffile", required_argument, NULL, 'c'},
		{"cnf-file", required_argument, NULL, 'c'},
		{"input", required_argument, NULL, 'i'},
		{"dont-use-nntp", no_argument, &dont_use_nntp, 1},
		{"dont-use-smtp", no_argument, &dont_use_smtp, 1},
		{"only-mail", no_argument, &work_only, ONLY_MAIL},
		{"only-news", no_argument, &work_only, ONLY_NEWS},
		{"only-special", no_argument, &work_only, ONLY_SPECIAL},
		{"only-convert", no_argument, 0, 'O'},
		{NULL, 0, NULL, 0}
	};
	int option_index = 0;
	const char *slash;

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

	parse_long_options (argc, argv, program_name, PACKAGE_VERSION, usage1);

    while (1)
    {
        int c;

        c = getopt_long (argc, argv, "i:Oc:", cmdlinopts, &option_index);

        /* Detect the end of the options.   */
        if (c == -1)
            break;
        switch (c)
        {
        default:
            break;
        case 'h':
            usage (0,NULL);
            break;
        case 'i':
			inputfilename=optarg;
            break;
        case 'c':
            cnffname = optarg;
            break;
        case 'O':
            argument_only_convert = 1;
            break;
        case '?':
            log (LOG_ERR, _("unknown option %s"), argv[optind - 1]);
            usage (2,_("unknown option"));
            break;
        }
    }                           /* while(1) */

    if (!argv[optind])
    {
        log (LOG_ERR, _("channel not given"));
        exit(2);
    }
    else if (argv[optind+1])
    {
        log (LOG_ERR, _("to many arguments"));
        exit (2);
    }
    action=argv[optind];

    /* configfile lesen */
    if (do_conf (cnffname, argv[optind]))
        exit (1);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));
	open_msglog();

	if (!dom_init())
		exit(1);
	if (info_read_list())
		exit(1);
	if (!grp_load_list(TRUE))
		exit (1);

#if defined(WANT_NNTP)
	if (opts.newshost)
		nntp_inquire_open(opts.newshost);
#endif

	exitcode=!maus2rfc (inputfilename);

#if defined(WANT_NNTP)
	if (opts.newshost)
		nntp_inquire_close();
#endif

	if (!grp_save_list())
		log(LOG_ERR, _("cannot save grouplist"));

	if (new_status_f)
		send_new_status();

	srfcmail_end_all();
	srfcnews_end_all();

	info_write_list();

	closelog ();
	close_msglog();

	exit (exitcode);
}

